"""
Author: Wenyu Ouyang
Date: 2023-10-25 17:12:30
LastEditTime: 2025-01-08 11:10:46
LastEditors: Wenyu Ouyang
Description: Reading public reservoir datasets
FilePath: \hydrodatasource\hydrodatasource\reader\reservoir_datasets.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import collections
import geopandas as gpd
from hydrodatasource.reader.data_source import HydroData


class Crd(HydroData):
    """Reading CRD reservoir dataset"""

    def __init__(self, data_path):
        self.data_source_dir = data_path
        self.data_source_description = self.set_data_source_describe()
        self.reservoir_info = self.read_reservoir_info()

    def get_name(self):
        return "CRD"

    def set_data_source_describe(self):
        data_root_dir = self.data_source_dir
        all_rsvrs_dir = os.path.join(data_root_dir, "CRD_v11_all_reservoirs")
        record_rsvrs_dir = os.path.join(data_root_dir, "CRD_v11_record_reservoirs")
        return collections.OrderedDict(
            ALL_RSVRS_SHPFILE=os.path.join(all_rsvrs_dir, "CRD_v11_all_reservoirs.shp"),
            RECORD_RSVRS_SHPFILE=os.path.join(
                record_rsvrs_dir, "CRD_v11_record_reservoirs.shp"
            ),
            # link: https://mp.weixin.qq.com/s/Iy3fFCtGkNc_iu0Ggkh1Lg
        )

    def read_reservoir_info(self):
        all_rsvrs_shpfile = self.data_source_description["ALL_RSVRS_SHPFILE"]
        all_rsvrs_shp = gpd.read_file(all_rsvrs_shpfile)
        # add unit metadata
        all_rsvrs_shp.attrs["units"] = {
            "Area": "km^2",
            "STOR": "km^3",
            "DIS_AV_CMS": "m^3/s",
            "Shape_Leng": "degree",  # Shape_Leng unit is degree (lat lon)
            "Shape_Area": "degree^2",
        }
        return all_rsvrs_shp



"""
Author: Wenyu Ouyang
Date: 2023-10-31 09:26:31
LastEditTime: 2025-01-09 19:15:20
LastEditors: Wenyu Ouyang
Description: Reading cleaned reservoir inflow data
FilePath: \hydrodatasource\hydrodatasource\reader\rsvr_inflow_reader.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import collections
import os

import pandas as pd
from tqdm import tqdm
import xarray as xr
from hydrodatasource.reader.data_source import HydroData


class RsvrInflowReader(HydroData):
    def __init__(self, data_folder):
        self.data_source_dir = data_folder
        self.data_source_description = self.set_data_source_describe()
        self.rsvr_info = self.read_rsvr_info()

    def set_data_source_describe(self):
        return collections.OrderedDict(
            {
                "RSVR_INFO_FILE": os.path.join(self.data_source_dir, "rsvr_info.csv"),
            }
        )

    def read_rsvr_info(self):
        return pd.read_csv(
            self.data_source_description["RSVR_INFO_FILE"], dtype={"STCD": str}
        )

    def read_rsvr_inflow(self):
        # Create an empty xarray.Dataset
        rsvr_inflow_data = xr.Dataset()

        # Iterate over all reservoir IDs in rsvr_info with tqdm progress bar
        inflow_data_list = []
        for rsvr_id in tqdm(
            self.rsvr_info["STCD"], desc="Reading reservoir inflow data"
        ):
            # Read inflow data for a single reservoir
            single_rsvr_inflow = self.read_1rsvr_inflow(rsvr_id)
            inflow_data_list.append(single_rsvr_inflow)
        # check the earlyest date
        start_date = pd.Timestamp.max
        for inflow in inflow_data_list:
            if inflow.index[0] < start_date:
                start_date = inflow.index[0]
        # if the start date is too early (before 1980), we will drop the data before 1980
        if start_date.year < 1980:
            start_date = pd.Timestamp(year=1980, month=1, day=1)
        # Concatenate all inflow data along the 'STCD' dimension
        inflow_data = xr.concat(
            [
                xr.DataArray(
                    inflow.loc[start_date:]["INQ"].values,
                    dims=["time"],
                    coords={
                        "time": inflow.loc[start_date:].index.values,
                        "STCD": rsvr_id,
                    },
                )
                for inflow, rsvr_id in zip(inflow_data_list, self.rsvr_info["STCD"])
            ],
            dim="STCD",
        )

        # Add the inflow data to the xarray.Dataset
        rsvr_inflow_data["inflow"] = inflow_data

        return rsvr_inflow_data

    def read_1rsvr_inflow(self, rsvr_id):
        """
        Read inflow data for a single reservoir.

        Parameters
        ----------
        rsvr_id : str
            Reservoir ID

        Returns
        -------
        pd.DataFrame
            A DataFrame containing inflow data for the specified reservoir.
        """
        file_path = os.path.join(
            self.data_source_dir, rsvr_id, f"{rsvr_id}_rsvr_data.csv"
        )
        df = pd.read_csv(
            file_path, parse_dates=["TM"], dtype={"STCD": str, "INQ": float}
        )
        df.set_index("TM", inplace=True)

        return df



import collections
import glob
import json
import os
import re
import numpy as np
import pandas as pd
import xarray as xr
from abc import ABC
from tqdm import tqdm
import polars as pl

from hydroutils import hydro_file
from hydroutils.hydro_time import generate_start0101_time_range

import hydrodatasource.configs.config as conf
from hydrodatasource.configs.config import CACHE_DIR
from hydrodatasource.configs.data_consts import (
    ERA5LAND_ET_REALATED_VARS,
    MODIS_ET_PET_8D_VARS,
)
from hydrodatasource.utils.utils import (
    cal_area_from_shp,
    calculate_basin_offsets,
    is_minio_folder,
    minio_file_list,
)
from hydrodatasource.reader import access_fs
import geopandas as gpd
from tqdm import tqdm


class HydroData(ABC):
    """An interface for reading multi-modal data sources.

    Parameters
    ----------
    ABC : _type_
        _description_
    """

    def __init__(self, data_path, dataset_name):
        self.data_source_dir = os.path.join(data_path, dataset_name)
        self.dataset_name = dataset_name

    def get_name(self):
        raise NotImplementedError

    def set_data_source_describe(self):
        raise NotImplementedError

    def read_data(self):
        raise NotImplementedError


class SelfMadeHydroDataset(HydroData):
    """A class for reading hydrodataset, but not really ready-datasets,
    just some data directorys organized like a HydroDataset.

    NOTE:
    We compile forcing data and attr data into a directory,
    organized like a ready dataset -- like Caravan.
    Only two directories are needed: attributes and timeseries
    """

    def __init__(self, data_path, dataset_name, time_unit=None, **kwargs):
        """Initialize a self-made Caravan-style dataset.

        Parameters
        ----------
        data_path : str
            The path to the custom-made data sources' parent directory.
        dataset_name : str
            SelfMadeHydroDataset's name, for example, googleflood or fdsources,
            different dataset may use this same datasource class, but they have different dataset_name.
        time_unit : list, optional
            we have different time units, by default None
        kwargs : dict, optional
            additional keyword arguments, by default None
        """
        if time_unit is None:
            time_unit = ["1D"]
        if any(unit not in ["1h", "3h", "1D", "8D"] for unit in time_unit):
            raise ValueError(
                "time_unit must be one of ['1h', '3h', '1D', '8D']. We only support these time units now."
            )
        # TODO: maybe starting with "s3://" is a better idea?
        self.head = "minio" if "s3://" in data_path else "local"
        super().__init__(data_path, dataset_name)
        self.data_source_description = self.set_data_source_describe()
        self.camels_sites = self.read_site_info()
        self.time_unit = time_unit
        # version is used for the version of the dataset, for example, camels_v2.0
        self.version = kwargs.get("version", None)
        # offset_to_utc is used for the offset to UTC, for example, for Chinese basins' data, we generally set it to True as we always use 08:00 with Beijing Time
        self.offset_to_utc = kwargs.get("offset_to_utc", None)
        self.trange4cache = kwargs.get("trange4cache", None)

    @property
    def streamflow_unit(self):
        unit_mapping = {"1h": "mm/h", "3h": "mm/3h", "1D": "mm/d"}
        return {unit: unit_mapping[unit] for unit in self.time_unit}

    def get_name(self):
        return "SelfMadeHydroDataset"

    def set_data_source_describe(self):
        data_root_dir = self.data_source_dir
        ts_dir = os.path.join(data_root_dir, "timeseries")
        # we assume that each subdirectory in ts_dir represents a time unit
        # In this subdirectory, there are csv files for each basin
        time_units_dir = self._where_ts_dir(ts_dir)
        pattern = os.path.join(ts_dir, "*_units_info.json")
        unit_files = glob.glob(pattern)
        attr_dir = os.path.join(data_root_dir, "attributes")
        attr_file = os.path.join(attr_dir, "attributes.csv")
        shape_dir = os.path.join(data_root_dir, "shapes")

        return collections.OrderedDict(
            DATA_DIR=data_root_dir,
            TS_DIRS=time_units_dir,
            ATTR_DIR=attr_dir,
            ATTR_FILE=attr_file,
            UNIT_FILES=unit_files,
            SHAPE_DIR=shape_dir,
        )

    def _where_ts_dir(self, ts_dir):
        return (
            [
                os.path.join(ts_dir, name)
                for name in minio_file_list(ts_dir)
                if is_minio_folder(os.path.join(ts_dir, name))
            ]
            if "s3://" in ts_dir
            else [
                os.path.join(ts_dir, name)
                for name in os.listdir(ts_dir)
                if os.path.isdir(os.path.join(ts_dir, name))
            ]
        )

    def read_site_info(self):
        camels_file = self.data_source_description["ATTR_FILE"]
        attrs = access_fs.spec_path(camels_file, head=self.head)
        return attrs[["basin_id", "area"]]

    def read_object_ids(self, object_params=None) -> np.array:
        return self.camels_sites["basin_id"].values

    def _validate_time_alignment(self, date, time_unit, start_hour_in_a_day, object_id):
        """
        Validate that data time alignment matches the expected start_hour_in_a_day.

        Parameters
        ----------
        date : np.ndarray
            Array of datetime values from the data
        time_unit : str
            Time unit string (e.g., "3h", "1D")
        start_hour_in_a_day : int
            Expected start hour in a day (0-23)
        object_id : str
            Basin/object ID for error reporting

        Raises
        ------
        ValueError
            If data alignment does not match the expected start_hour_in_a_day
        """
        if len(date) == 0:
            return

        # Check if this is a sub-daily interval (contains 'h' or 'H')
        if "h" not in time_unit.lower() and "H" not in time_unit:
            return

        # Extract the numeric part to check if it's not 1h
        numeric_part = "".join(filter(str.isdigit, time_unit))
        if not numeric_part or int(numeric_part) == 1:
            return

        # For intervals like 3h, validate alignment
        expected_hour = start_hour_in_a_day
        # Get actual hours from the data
        actual_hours = pd.to_datetime(date).hour
        unique_hours = np.unique(actual_hours)

        # Calculate expected hours based on the interval
        interval_hours = int(numeric_part)
        expected_hours_in_day = list(range(expected_hour, 24, interval_hours))

        # Check if any actual hour is not in expected hours
        misaligned_hours = [h for h in unique_hours if h not in expected_hours_in_day]

        if misaligned_hours:
            # Find what the correct start_hour should be
            actual_start_hour = min(unique_hours)
            raise ValueError(
                f"Data time alignment error for basin '{object_id}': "
                f"The actual data starts at hour {actual_start_hour:02d}, "
                f"but start_hour_in_a_day is set to {expected_hour}. "
                f"Expected hours in a day: {sorted(expected_hours_in_day)}, "
                f"but found hours: {sorted(unique_hours.tolist())}. "
                f"Please set start_hour_in_a_day to {actual_start_hour} to match your data."
            )

    def read_timeseries(
        self, object_ids=None, t_range_list: list = None, relevant_cols=None, **kwargs
    ) -> dict:
        """
        Returns a dictionary containing data with different time scales.

        Parameters
        ----------
        object_ids : list, optional
            List of object IDs. Defaults to None.
        t_range_list : list, optional
            List of time ranges. Defaults to None.
        relevant_cols : list, optional
            List of relevant columns. Defaults to None.
        **kwargs : dict, optional
            Additional keyword arguments.
            time_units : list, optional
                List of time units to process
            start0101_freq : bool, optional
                For freq setting, if the start date is 01-01, set True
            offset_to_utc : bool, optional
                Whether to offset the time to UTC
            start_hour_in_a_day : int, optional
                The start hour in a day for sub-daily intervals (0-23). Default is 2.

        Returns
        -------
        dict
            A dictionary containing data with different time scales.
        """
        time_units = kwargs.get("time_units", ["1D"])
        start0101_freq = kwargs.get("start0101_freq", False)
        offset_to_utc = kwargs.get("offset_to_utc", self.offset_to_utc)
        start_hour_in_a_day = kwargs.get("start_hour_in_a_day", 2)

        # Validate start_hour_in_a_day range
        if not isinstance(start_hour_in_a_day, int) or not (
            0 <= start_hour_in_a_day <= 23
        ):
            raise ValueError(
                f"start_hour_in_a_day must be an integer between 0 and 23, got {start_hour_in_a_day}"
            )

        results = {}

        for time_unit in time_units:
            if offset_to_utc:
                basinoutlets_path = os.path.join(
                    self.data_source_description["SHAPE_DIR"], "basinoutlets.shp"
                )
                try:
                    offset_dict = calculate_basin_offsets(basinoutlets_path)
                except FileNotFoundError as e:
                    raise FileNotFoundError(
                        f"basinoutlets.shp not found in {basinoutlets_path}."
                    ) from e
            ts_dir = self._get_ts_dir(
                self.data_source_description["TS_DIRS"], time_unit
            )
            if start0101_freq:
                t_range = generate_start0101_time_range(
                    start_time=t_range_list[0],
                    end_time=t_range_list[-1],
                    freq=time_unit,
                )
            else:
                t_range = pd.date_range(
                    start=t_range_list[0], end=t_range_list[-1], freq=time_unit
                )
            nt = len(t_range)
            x = np.full([len(object_ids), nt, len(relevant_cols)], np.nan)

            # Flag to check data alignment only once
            data_alignment_checked = False

            for k in tqdm(
                range(len(object_ids)), desc=f"Reading timeseries data with {time_unit}"
            ):
                ts_file = os.path.join(
                    ts_dir,
                    object_ids[k] + ".csv",
                )
                if "s3://" in ts_file:
                    with conf.FS.open(ts_file, mode="rb") as f:
                        ts_data = pd.read_csv(f, engine="c")
                else:
                    ts_data = pd.read_csv(ts_file, engine="c")
                date = pd.to_datetime(ts_data["time"]).values
                if offset_to_utc:
                    date = date - np.timedelta64(offset_dict[object_ids[k]], "h")

                # Validate data alignment with start_hour_in_a_day (only check once)
                if not data_alignment_checked:
                    self._validate_time_alignment(
                        date, time_unit, start_hour_in_a_day, object_ids[k]
                    )
                    data_alignment_checked = True

                [_, ind1, ind2] = np.intersect1d(date, t_range, return_indices=True)

                for j in range(len(relevant_cols)):
                    tmp_ = self._read_timeseries_1basin1var(ts_data, relevant_cols[j])
                    x[k, ind2, j] = tmp_[ind1]

            results[time_unit] = x

        return results

    def _read_timeseries_1basin1var(self, ts_data, relevant_col):
        if "precipitation" in relevant_col:
            prcp = ts_data[relevant_col].values
            prcp[prcp < 0] = 0.0
            return prcp
        elif relevant_col in ERA5LAND_ET_REALATED_VARS:
            evap = -1 * ts_data[relevant_col].values
            evap[evap < 0] = 0.0
            return evap
        elif relevant_col in MODIS_ET_PET_8D_VARS:
            modis_values = ts_data[relevant_col].values
            modis_dates = pd.to_datetime(ts_data["time"].values)
            for idx, current_date in enumerate(modis_dates):
                # Check if the date is prior to or on January 1st
                if current_date.month == 1 and current_date.day == 1:
                    if idx == 0:
                        # First day is January 1st, no previous date to scale from
                        continue
                        # Get the previous date
                    previous_date = modis_dates[idx - 1]
                    # Calculate the number of days between the previous date and January 1st
                    delta_days = (current_date - previous_date).days

                    # Adjust the MODIS value based on the number of days between the previous date and January 1st
                    if delta_days > 0:
                        modis_values[idx - 1] = modis_values[idx - 1] * 8 / delta_days
            # NOTE: MODIS ET values are ACTUALLY in 0.1mm/day, so we need to convert to mm/day
            return modis_values * 0.1
        else:
            return ts_data[relevant_col].values

    def read_attributes(
        self, object_ids=None, constant_cols=None, **kwargs
    ) -> np.array:
        """2d data (site_num * var_num), non-time-series data"""
        attr_file = self.data_source_description["ATTR_FILE"]
        if "s3://" in attr_file:
            with conf.FS.open(attr_file, mode="rb") as f:
                attrs = pd.read_csv(f, dtype={"basin_id": str})
        else:
            attrs = pd.read_csv(attr_file, dtype={"basin_id": str})
        if object_ids is None:
            if constant_cols is None:
                return attrs
            object_ids = attrs["basin_id"].values
        if constant_cols is None:
            constant_cols = attrs.columns.values
        x = np.full([len(object_ids), len(constant_cols)], np.nan)
        for k in range(len(object_ids)):
            ind = attrs["basin_id"] == object_ids[k]
            for j in range(len(constant_cols)):
                x[k, j] = attrs[constant_cols[j]][ind].values
        return x

    def get_attributes_cols(self) -> np.array:
        """the constant cols in this data_source"""
        attr_file = self.data_source_description["ATTR_FILE"]
        if "s3://" in attr_file:
            with conf.FS.open(attr_file, mode="rb") as f:
                attrs = pd.read_csv(f, dtype={"basin_id": str})
        else:
            attrs = pd.read_csv(attr_file, dtype={"basin_id": str})
        attr_units = attrs.columns.values
        return self._check_vars_in_unitsinfo(attr_units)

    def get_timeseries_cols(self) -> np.array:
        """the relevant cols in this data_source"""
        ts_dirs = self.data_source_description["TS_DIRS"]
        unit_files = self.data_source_description["UNIT_FILES"]
        all_vars = {}
        for time_unit in self.time_unit:
            # Find the directory that corresponds to the current time unit
            ts_dir = self._get_ts_dir(ts_dirs, time_unit)
            # Find the corresponding unit file
            unit_file = next(
                file
                for file in unit_files
                if f"{time_unit}_units_info.json" == file.split(os.sep)[-1]
            )
            # Load the first CSV file in the directory to extract column names
            if "s3://" in ts_dir:
                ts_file = os.path.join(ts_dir, minio_file_list(ts_dir)[0])
                with conf.FS.open(ts_file, mode="rb") as f:
                    ts_tmp = pd.read_csv(f, dtype={"basin_id": str})
            else:
                ts_file = os.path.join(ts_dir, os.listdir(ts_dir)[0])
                ts_tmp = pd.read_csv(ts_file, dtype={"basin_id": str})
            # Get the relevant forcing units and validate against unit info
            forcing_units = ts_tmp.columns.values[1:]
            the_vars = self._check_vars_in_unitsinfo(forcing_units, unit_file)
            # Map the variables to the corresponding time unit
            all_vars[time_unit] = the_vars
        return all_vars

    def _get_ts_dir(self, ts_dirs, time_unit):
        """we add version for ts directory, so we need to find the correct ts directory

        Parameters
        ----------
        ts_dirs : list
            the list of ts directories without version
        time_unit : str
            the time unit

        Returns
        -------
        _type_
            _description_
        """
        ts_dir = next(
            dir_path for dir_path in ts_dirs if time_unit == dir_path.split(os.sep)[-1]
        )
        version = self.version
        ts_dir = (
            ts_dir + f"_{version}" if version is not None and version != "" else ts_dir
        )

        return ts_dir

    def _check_vars_in_unitsinfo(self, vars, unit_file=None):
        """If a var is not recorded in a units_info file, we will not use it.

        Parameters
        ----------
        vars : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        if unit_file is None:
            # For attributes, all the variables' units are same in all unit_info files
            # hence, we just chose the first one
            unit_file = self.data_source_description["UNIT_FILES"][0]
        if "s3://" in unit_file:
            with conf.FS.open(unit_file, mode="rb") as fp:
                units_info = json.load(fp)
        else:
            units_info = hydro_file.unserialize_json(unit_file)
        vars_final = [var_ for var_ in vars if var_ in units_info]
        return np.array(vars_final)

    def cache_attributes_xrdataset(self):
        """Convert all the attributes to a single dataset

        Returns
        -------
        None
        """
        # NOTICE: although it seems that we don't use pint_xarray, we have to import this package
        import pint_xarray  # noqa: F401

        shape_dir = os.path.join(
            self.data_source_description["SHAPE_DIR"], "basins.shp"
        )
        if "s3://" in shape_dir:
            with conf.FS.open(shape_dir, mode="rb") as f:
                shape = gpd.read_file(f)
        else:
            shape = gpd.read_file(shape_dir)
        df_area = cal_area_from_shp(shape)  # calculate the area from shape file
        df_area.set_index("basin_id", inplace=True)

        df_attr = self.read_attributes()
        df_attr.set_index("basin_id", inplace=True)
        df_attr = df_attr.join(df_area)
        # Mapping provided units to the variables in the datasets
        # For attributes, all the variables' units are same in all unit_info files
        # hence, we just chose the first one
        if "s3://" in self.data_source_description["UNIT_FILES"][0]:
            with conf.FS.open(
                self.data_source_description["UNIT_FILES"][0], mode="rb"
            ) as fp:
                units_dict = json.load(fp)
        else:
            units_dict = hydro_file.unserialize_json(
                self.data_source_description["UNIT_FILES"][0]
            )
        units_dict["shp_area"] = "km^2"  # add the unit of shp_area
        units_dict["area"] = "km^2"  # add the unit of area
        # Convert string columns to categorical variables and record categorical mappings
        categorical_mappings = {}
        for column in df_attr.columns:
            if df_attr[column].dtype == "object":
                df_attr[column] = df_attr[column].astype("category")
                categorical_mappings[column] = dict(
                    enumerate(df_attr[column].cat.categories)
                )
                df_attr[column] = df_attr[column].cat.codes

        ds = xr.Dataset()
        for column in df_attr.columns:
            attrs = {"units": units_dict.get(column, "unknown")}
            if column in categorical_mappings:
                attrs["category_mapping"] = categorical_mappings[column]

            data_array = xr.DataArray(
                data=df_attr[column].values,
                dims=["basin"],
                # we have set gage_id as index so that it won't be saved as numeric values
                coords={"basin": df_attr.index.values.astype(str)},
                attrs=attrs,
            )
            ds[column] = data_array

        # Convert categorical mappings to strings
        for column in ds.data_vars:
            if "category_mapping" in ds[column].attrs:
                # Convert the dictionary to a string
                mapping_str = str(ds[column].attrs["category_mapping"])
                ds[column].attrs["category_mapping"] = mapping_str
        dataset_name = self.dataset_name
        prefix_ = "" if dataset_name is None else dataset_name + "_"
        ds.to_netcdf(os.path.join(CACHE_DIR, f"{prefix_}attributes.nc"))

    def cache_timeseries_xrdataset(self, **kwargs):
        """Save all timeseries data in separate NetCDF files for each time unit.

        Parameters
        ----------
        t_range : list, optional
            Time range for the data, by default ["1980-01-01", "2023-12-31"]
        kwargs : dict, optional
            batchsize -- Number of basins to process per batch, by default 100
            time_units -- List of time units to process, by default None
            start0101_freq -- for freq setting, if the start date is 01-01, set True, by default False
            offset_to_utc -- whether to offset the time to UTC, by default False
            start_hour_in_a_day -- the start hour in a day (0-23), by default 2 which means 2-5-8-11-14-17-20-23 UTC.
                                   Chinese basins data always use 08:00 with Beijing Time, so we set the default value to 2.
                                   Only applicable for sub-daily intervals (currently only "3h" is supported)
        """
        batchsize = kwargs.get("batchsize", 100)
        time_units = kwargs.get("time_units", self.time_unit) or [
            "1D"
        ]  # Default to ["1D"] if not specified or if time_units is None
        start0101_freq = kwargs.get("start0101_freq", False)
        offset_to_utc = kwargs.get("offset_to_utc", self.offset_to_utc)
        start_hour_in_a_day = kwargs.get("start_hour_in_a_day", 2)

        # Validate start_hour_in_a_day
        if not isinstance(start_hour_in_a_day, int) or not (
            0 <= start_hour_in_a_day <= 23
        ):
            raise ValueError(
                f"start_hour_in_a_day must be an integer between 0 and 23, got {start_hour_in_a_day}"
            )
        variables = self.get_timeseries_cols()
        basins = self.camels_sites["basin_id"].values

        # Validate time_units for sub-daily intervals
        for time_unit in time_units:
            # Check if this is an hourly interval (contains 'h' or 'H')
            if "h" in time_unit.lower() or "H" in time_unit:
                # Extract the numeric part
                numeric_part = "".join(filter(str.isdigit, time_unit))
                if numeric_part and int(numeric_part) != 1:
                    # If it's not 1h, only allow 3h
                    if time_unit.lower() not in ["3h"]:
                        raise ValueError(
                            f"Currently only '3h' sub-daily interval is supported for custom start_hour_in_a_day. "
                            f"Got '{time_unit}'. Please use '1h', '3h', or daily/longer intervals."
                        )

        # Define the generator function for batching
        def data_generator(basins, batch_size):
            for i in range(0, len(basins), batch_size):
                yield basins[i : i + batch_size]

        for time_unit in time_units:
            if self.trange4cache is None:
                if time_unit != "3h":
                    self.trange4cache = ["1960-01-01", "2024-12-31"]
                else:
                    # Calculate the end hour based on 3-hour intervals
                    # For 3h intervals, find the last timestamp within a day
                    # Example: start_hour_in_a_day=2 -> 02, 05, 08, 11, 14, 17, 20, 23 (last is 23)
                    # Example: start_hour_in_a_day=5 -> 05, 08, 11, 14, 17, 20, 23 (last is 23)
                    start_hour = str(start_hour_in_a_day).zfill(2)
                    # Find the last hour in the day for this interval
                    hours_in_day = list(range(start_hour_in_a_day, 24, 3))
                    if len(hours_in_day) > 0:
                        end_hour = str(hours_in_day[-1]).zfill(2)
                    else:
                        # If start_hour >= 24, which shouldn't happen, default to 23
                        end_hour = "23"
                    self.trange4cache = [
                        f"1960-01-01 {start_hour}",
                        f"2024-12-31 {end_hour}",
                    ]

            # Generate the time range specific to the time unit
            if start0101_freq:
                times = (
                    generate_start0101_time_range(
                        start_time=self.trange4cache[0],
                        end_time=self.trange4cache[-1],
                        freq=time_unit,
                    )
                    .strftime("%Y-%m-%d %H:%M:%S")
                    .tolist()
                )
            else:
                times = (
                    pd.date_range(
                        start=self.trange4cache[0],
                        end=self.trange4cache[-1],
                        freq=time_unit,
                    )
                    .strftime("%Y-%m-%d %H:%M:%S")
                    .tolist()
                )
            # Retrieve the correct units information for this time unit
            unit_file = next(
                file
                for file in self.data_source_description["UNIT_FILES"]
                if time_unit in file
            )
            if "s3://" in unit_file:
                with conf.FS.open(unit_file, mode="rb") as fp:
                    units_info = json.load(fp)
            else:
                units_info = hydro_file.unserialize_json(unit_file)

            for basin_batch in data_generator(basins, batchsize):
                data = self.read_timeseries(
                    object_ids=basin_batch,
                    t_range_list=self.trange4cache,
                    relevant_cols=variables[
                        time_unit
                    ],  # Ensure we use the right columns for the time unit
                    time_units=[
                        time_unit
                    ],  # Pass the time unit to ensure correct data retrieval
                    start0101_freq=start0101_freq,
                    offset_to_utc=offset_to_utc,
                    start_hour_in_a_day=start_hour_in_a_day,
                )

                dataset = xr.Dataset(
                    data_vars={
                        variables[time_unit][i]: (
                            ["basin", "time"],
                            data[time_unit][:, :, i],
                            {"units": units_info[variables[time_unit][i]]},
                        )
                        for i in range(len(variables[time_unit]))
                    },
                    coords={
                        "basin": basin_batch,
                        "time": pd.to_datetime(times),
                    },
                )

                # Save the dataset to a NetCDF file for the current batch and time unit
                prefix_ = self._get_ts_file_prefix_(self.dataset_name, self.version)
                batch_file_path = os.path.join(
                    CACHE_DIR,
                    f"{prefix_}timeseries_{time_unit}_batch_{basin_batch[0]}_{basin_batch[-1]}.nc",
                )
                dataset.to_netcdf(batch_file_path)

                # Release memory by deleting the dataset
                del dataset
                del data

    def cache_xrdataset(self, t_range=None, time_units=None):
        """Save all data in a netcdf file in the cache directory"""
        self.cache_attributes_xrdataset()
        self.cache_timeseries_xrdataset(time_units=time_units)

    def read_ts_xrdataset(
        self,
        gage_id_lst: list = None,
        t_range: list = None,
        var_lst: list = None,
        **kwargs,
    ) -> dict:
        """
        Read time-series xarray dataset from multiple NetCDF files and organize them by time units.

        Parameters:
        ----------
        gage_id_lst: list
            List of gage IDs to select.
        t_range: list
            List of two elements [start_time, end_time] to select time range.
        var_lst: list
            List of variables to select.
        **kwargs
            Additional arguments.

        Returns:
        ----------
        dict: A dictionary where each key is a time unit and each value is an xarray.Dataset containing the selected gage IDs, time range, and variables.
        """
        dataset_name = self.dataset_name
        version = self.version
        time_units = kwargs.get("time_units", self.time_unit)
        recache = kwargs.get("recache", False)
        if var_lst is None:
            return None

        # Initialize a dictionary to hold datasets for each time unit
        datasets_by_time_unit = {}

        prefix_ = self._get_ts_file_prefix_(dataset_name, version)

        for time_unit in time_units:
            # Collect batch files specific to the current time unit
            batch_files = self._get_batch_files(prefix_, time_unit)

            if not batch_files or recache:
                # Cache the data if no batch files are found for the current time unit
                self.cache_timeseries_xrdataset(**kwargs)
                batch_files = self._get_batch_files(prefix_, time_unit)

            selected_datasets = []

            for batch_file in batch_files:
                ds = xr.open_dataset(batch_file)
                all_vars = ds.data_vars
                if any(var not in ds.variables for var in var_lst):
                    raise ValueError(f"var_lst must all be in {all_vars}")
                if valid_gage_ids := [
                    gid for gid in gage_id_lst if gid in ds["basin"].values
                ]:
                    ds_selected = ds[var_lst].sel(
                        basin=valid_gage_ids, time=slice(t_range[0], t_range[1])
                    )
                    selected_datasets.append(ds_selected)

                ds.close()  # Close the dataset to free memory

            # If any datasets were selected, concatenate them along the 'basin' dimension
            if selected_datasets:
                # NOTE: the chosen part must be sorted by basin, or there will be some negative sideeffect for continue usage of this repo
                datasets_by_time_unit[time_unit] = xr.concat(
                    selected_datasets, dim="basin"
                ).sortby("basin")
            else:
                datasets_by_time_unit[time_unit] = xr.Dataset()

        return datasets_by_time_unit

    def _get_ts_file_prefix_(self, dataset_name, version):
        prefix_ = "" if dataset_name is None else dataset_name + "_"
        # we add version for prefix_ as we will update the dataset iteratively
        prefix_ = prefix_ + f"{version}_" if version is not None else prefix_
        return prefix_

    def _get_batch_files(self, prefix_, time_unit):
        return [
            os.path.join(CACHE_DIR, f)
            for f in os.listdir(CACHE_DIR)
            if re.match(
                rf"^{prefix_}timeseries_{time_unit}_batch_[A-Za-z0-9_]+_[A-Za-z0-9_]+\.nc$",
                f,
            )
        ]

    def read_attr_xrdataset(self, gage_id_lst=None, var_lst=None, **kwargs):
        dataset_name = self.dataset_name

        prefix_ = "" if dataset_name is None else dataset_name + "_"
        if var_lst is None or len(var_lst) == 0:
            return None
        recache = kwargs.get("recache", False)
        if (
            not os.path.exists(os.path.join(CACHE_DIR, f"{prefix_}attributes.nc"))
            or recache
        ):
            self.cache_attributes_xrdataset()
        attr = xr.open_dataset(os.path.join(CACHE_DIR, f"{prefix_}attributes.nc"))
        return attr[var_lst].sel(basin=gage_id_lst)

    def read_area(self, gage_id_lst=None):
        """read area of each basin/unit"""
        return self.read_attr_xrdataset(gage_id_lst, ["area"])

    def read_mean_prcp(self, gage_id_lst=None, unit="mm/d"):
        """read mean precipitation of each basin
        default unit is mm/d, but one can chose other units and we will convert the unit to the specified unit

        Parameters
        ----------
        gage_id_lst : list, optional
            the list of gage ids, by default None
        unit : str, optional
            the unit of precipitation, by default "mm/d"

        Returns
        -------
        xr.Dataset
            the mean precipitation of each basin
        """
        pre_mm_syr = self.read_attr_xrdataset(gage_id_lst, ["pre_mm_syr"])
        da = pre_mm_syr["pre_mm_syr"]
        # Convert the unit to the specified unit, pre_mm_syr means yearly precipitation
        if unit in ["mm/d", "mm/day"]:
            converted_data = da / 365
        elif unit in ["mm/h", "mm/hour"]:
            converted_data = da / 8760
        elif unit in ["mm/3h", "mm/3hour"]:
            converted_data = da / (8760 / 3)
        elif unit in ["mm/8d", "mm/8day"]:
            converted_data = da / (365 / 8)
        else:
            raise ValueError(
                "unit must be one of ['mm/d', 'mm/day', 'mm/h', 'mm/hour', 'mm/3h', 'mm/3hour', 'mm/8d', 'mm/8day']"
            )

        # Set the units attribute
        converted_data.attrs["units"] = unit
        # Assign the modified DataArray back to the Dataset
        pre_mm_syr["pre_mm_syr"] = converted_data
        return pre_mm_syr


class SelfMadeForecastDataset(SelfMadeHydroDataset):
    """For selfmadehydrodataset, we design a new file format for forecast data from GFS et al."""

    def __init__(self, data_path, dataset_name, time_unit=None):
        """intialize a Class for reading forecast data

        Parameters
        ----------
        data_path : str
            the path of data source
        time_unit : list, optional
            unit of one time period, by default None
        dataset_name: str
            name will be used for cache files
        """
        super().__init__(data_path, dataset_name, time_unit)

    def set_data_source_describe(self):
        """set data source description

        Returns
        -------
        dict
            a dict with name and path of the data source
        """
        data_source_description = super().set_data_source_describe()
        forecast_dir = os.path.join(self.data_source_dir, "forecasts")
        forecast_ts_dir = self._where_ts_dir(forecast_dir)
        data_source_description["FORECAST_DIR"] = forecast_ts_dir
        return data_source_description

    def read_ts_xrdataset(
        self,
        gage_id_lst,
        t_range,
        var_lst,
        **kwargs,
    ):
        """读取预见期数据

        Parameters
        ----------
        gage_id_lst : list
            流域ID列表
        t_range : datetime
            time range [start_time, end_time]
        var_lst : list
            变量列表

        Returns
        -------
        xr.Dataset
            预见期数据
        """
        if forecast_mode := kwargs.get("forecast_mode", False):
            return self.read_forecast_xrdataset(
                gage_id_lst,
                t_range,
                var_lst,
                **kwargs,
            )
        else:
            return super(SelfMadeForecastDataset, self).read_ts_xrdataset(
                gage_id_lst,
                t_range,
                var_lst,
                **kwargs,
            )

    def read_forecast_xrdataset(
        self,
        gage_id_lst,
        t_range,
        var_lst,
        **kwargs,
    ):
        """read cache nc file

        Parameters
        ----------
        gage_id_lst : list
            the list of gage ids
        t_range : list
            the start time and end time
        variables : list
            variables list

        Returns
        -------
        xr.Dataset
            forecast data
        """
        # if None, we will just chose all lead time data
        lead_time = kwargs.get("lead_time", None)
        dataset_name = self.dataset_name + "_" + "forecast"
        version = self.version
        time_units = kwargs.get("time_units", self.time_unit)
        if var_lst is None:
            return None

        # Initialize a dictionary to hold datasets for each time unit
        datasets_by_time_unit = {}

        prefix_ = self._get_ts_file_prefix_(dataset_name, version)

        for time_unit in time_units:
            # Collect batch files specific to the current time unit
            batch_files = self._get_batch_files(prefix_, time_unit)

            if not batch_files:
                # Cache the data if no batch files are found for the current time unit
                self.cache_forecast_xrdataset(
                    variables=var_lst, time_units=[time_unit], prefix=prefix_
                )
                batch_files = self._get_batch_files(prefix_, time_unit)

            selected_datasets = []
            for batch_file in batch_files:
                ds = xr.open_dataset(batch_file)
                all_vars = ds.data_vars
                if any(var not in ds.variables for var in var_lst):
                    raise ValueError(f"var_lst must all be in {all_vars}")
                if valid_gage_ids := [
                    gid for gid in gage_id_lst if gid in ds["basin"].values
                ]:
                    ds_selected = ds[var_lst].sel(
                        basin=valid_gage_ids, time=slice(t_range[0], t_range[1])
                    )
                    selected_datasets.append(ds_selected)
                    if lead_time is None:
                        lead_time = ds["lead_step"].values
                ds.close()  # Close the dataset to free memory
            # If any datasets were selected, concatenate them along the 'basin' dimension
            if selected_datasets:
                # NOTE: the chosen part must be sorted by basin, or there will be some negative sideeffect for continue usage of this repo
                datasets_by_time_unit[time_unit] = xr.concat(
                    selected_datasets, dim="basin"
                )
            else:
                datasets_by_time_unit[time_unit] = xr.Dataset()
        return datasets_by_time_unit

    def cache_forecast_xrdataset(self, t_range=None, **kwargs):
        """Save all forecast data in separate NetCDF files for each batch of basins and time units.

        Parameters
        ----------
        t_range : list, optional
            Time range for the forecast_date, by default None
        kwargs : dict, optional
            batchsize -- Number of basins to process per batch, by default 100
            variables -- List of variables to process, by default None
            time_units -- List of time units to process, by default self.time_unit
            prefix -- Prefix for the NetCDF file names, by default self.dataset_name
        """
        batchsize = kwargs.get("batchsize", 100)
        variables = kwargs.get("variables", None)
        time_units = kwargs.get("time_units", self.time_unit)
        prefix_ = kwargs.get("prefix", self.dataset_name)

        # Get forecast directories (one for each time unit)
        forecast_dirs = self.data_source_description["FORECAST_DIR"]
        if isinstance(forecast_dirs, str):
            forecast_dirs = [forecast_dirs]

        # Process each time unit
        for time_unit, f_dir in zip(time_units, forecast_dirs):
            # Get time delta based on time_unit
            try:
                # 尝试从 time_unit 解析时间间隔
                time_delta = pd.Timedelta(time_unit)
            except ValueError:
                # 若解析失败，抛出不支持时间单位的异常
                raise ValueError(f"Unsupported time unit: {time_unit}")

            # Get all basin files for this time unit
            basin_files = [
                os.path.join(f_dir, f)
                for f in self._handle_file_operation(f_dir, "list")
                if f.endswith(".csv")
            ]
            # sort files by basin_id
            basin_files = sorted(
                basin_files, key=lambda x: os.path.splitext(os.path.basename(x))[0]
            )
            # Extract basin IDs from filenames
            basins = [os.path.splitext(os.path.basename(f))[0] for f in basin_files]

            # Define the generator function for batching
            def data_generator(basins, batch_size):
                for i in range(0, len(basins), batch_size):
                    yield basins[i : i + batch_size], basin_files[i : i + batch_size]

            # Process each batch
            for basin_batch, file_batch in data_generator(basins, batchsize):
                # Initialize data structure
                all_data = []
                all_times = []
                all_lead_steps = []

                # Read each file in the batch
                # 使用 tqdm 来显示进度条
                for basin_id, csv_file in tqdm(
                    zip(basin_batch, file_batch),
                    desc=f"Processing batch for {time_unit}",
                    total=len(basin_batch),
                ):
                    # Read CSV file
                    df = self._handle_file_operation(
                        csv_file, "read", parse_dates=["date", "forecast_date"]
                    )

                    # Filter by time range if provided
                    if t_range is not None:
                        mask = (df["forecast_date"] >= pd.to_datetime(t_range[0])) & (
                            df["forecast_date"] <= pd.to_datetime(t_range[-1])
                        )
                        df = df[mask]

                    # Calculate lead steps based on time difference
                    df["lead_step"] = (
                        (df["forecast_date"] - df["date"]) / time_delta
                    ).astype(int)

                    # Get unique times and lead steps
                    # 对日期和预见期步长进行排序
                    times = np.sort(df["date"].unique())
                    lead_steps = np.sort(df["lead_step"].unique())

                    # Fill data array using helper function
                    basin_data = self._fill_basin_data(df, times, lead_steps, variables)
                    all_data.append(basin_data)
                    all_times.append(times)
                    all_lead_steps.append(lead_steps)

                # Create xarray Dataset
                dataset = xr.Dataset(
                    data_vars={
                        variables[k]: (
                            ["basin", "time", "lead_step"],
                            np.stack([data[:, :, k] for data in all_data]),
                            {"units": "mm"},  # Adjust units as needed
                        )
                        for k in range(len(variables))
                    },
                    coords={
                        "basin": basin_batch,
                        "time": pd.to_datetime(all_times[0]),
                        "lead_step": all_lead_steps[0],
                    },
                )

                # Save the dataset to a NetCDF file with time_unit in filename
                batch_file_path = os.path.join(
                    CACHE_DIR,
                    f"{prefix_}timeseries_{time_unit}_batch_{basin_batch[0]}_{basin_batch[-1]}.nc",
                )
                dataset.to_netcdf(batch_file_path)

                # Release memory
                del dataset
                del all_data

    def _handle_file_operation(self, file_path, operation, mode="rb", **kwargs):
        """Handle file operations for both local and S3 paths uniformly.

        Parameters
        ----------
        file_path : str
            File path (local or S3).
        operation : str
            Type of operation ('list' to get the file list, 'read' to read the file).
        mode : str, optional
            File open mode, by default "rb".
        **kwargs : dict
            Other parameters (such as parse_dates, etc.).

        Returns
        -------
        Any
            Results returned according to the operation type.
        """
        if "s3://" in file_path:
            with conf.FS.open(file_path, mode=mode) as f:
                if operation == "list":
                    return minio_file_list(file_path)
                elif operation == "read":
                    return pd.read_csv(f, **kwargs)
        elif operation == "list":
            return os.listdir(file_path)
        elif operation == "read":
            return pd.read_csv(file_path, **kwargs)

    def _fill_basin_data(self, df, times, lead_steps, variables):
        """Fill the data array for a single basin

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame containing the raw data
        times : array-like
            Array of time points
        lead_steps : array-like
            Array of lead time steps
        variables : list
            List of variables

        Returns
        -------
        np.ndarray
            A filled 3D array (time × lead time × variables)
        """
        # Pre-allocate the result array
        basin_data = np.full((len(times), len(lead_steps), len(variables)), np.nan)

        # 使用pivot_table进行快速数据重组
        for k, var in enumerate(variables):
            if var in df.columns:
                # 使用pivot_table快速重组数据
                pivot_df = df.pivot_table(
                    values=var,
                    index="date",
                    columns="lead_step",
                    aggfunc="first",  # 取第一个值
                )

                # 确保pivot_df包含所有times和lead_steps
                pivot_df = pivot_df.reindex(index=times, columns=lead_steps)

                # 将数据填充到结果数组中
                if not pivot_df.empty:
                    basin_data[:, :, k] = pivot_df.values

        return basin_data

    def read_forecast(
        self, object_ids=None, t_range_list: list = None, relevant_cols=None, **kwargs
    ):
        """
        Read forecast data (hourly/daily forecasts) from CSV files, where each basin has its own file named after the basin_id.
        The time range in the parameters refers to the forecast_date (i.e., the target period of the forecast), not the date (the execution time of the forecast).

        Parameters
        ----------
        object_ids : list
            List of basin IDs.
        t_range_list : list
            Time range for the target forecast period [start_time, end_time].
        relevant_cols : list
            List of variable names to be read.
        Returns
        -------
        dict
            {basin_id: pd.DataFrame}, where each basin has a DataFrame filtered by the time range and variables.
        """
        if object_ids is None or t_range_list is None or relevant_cols is None:
            raise ValueError("object_ids, t_range_list, relevant_cols 不能为空")

        results = {}

        forecast_dirs = self.data_source_description.get("FORECAST_DIR", [])
        if isinstance(forecast_dirs, str):
            forecast_dirs = [forecast_dirs]

        for basin_id in object_ids:
            found = False
            for forecast_dir in forecast_dirs:
                csv_path = os.path.join(forecast_dir, f"{basin_id}.csv")
                if os.path.exists(csv_path):
                    found = True
                    break
            if not found:
                results[basin_id] = None
                continue

            df = pd.read_csv(csv_path, parse_dates=["date", "forecast_date"])
            mask = (df["forecast_date"] >= pd.to_datetime(t_range_list[0])) & (
                df["forecast_date"] <= pd.to_datetime(t_range_list[-1])
            )
            df = df.loc[mask]

            cols = ["date", "forecast_date"] + [
                col for col in relevant_cols if col in df.columns
            ]
            df = df[cols]

            results[basin_id] = df.reset_index(drop=True)

        return results


class StationHydroDataset(SelfMadeHydroDataset):
    """A class for reading hydrodataset with additional station data.

    This class extends SelfMadeHydroDataset to handle datasets that include
    a stations folder containing individual station data and basin-station
    relationship information.

    Directory structure:
    - attributes/
    - shapes/
    - timeseries/
    - stations/
      - 1D/                              # Daily data for all stations
      - 3h/                              # 3h data for all stations
      - basin_station_info/              # Basin-station relationship info
        - all_basin_station_mapping.csv
        - basin_summary.csv
        - basin_xxx_stations.csv
        - adjacency_xxx_True.csv
    """

    def __init__(self, data_path, dataset_name, time_unit=None, **kwargs):
        """Initialize StationHydroDataset.

        Parameters
        ----------
        data_path : str
            Path to the dataset directory
        dataset_name : str
            Name of the dataset
        time_unit : list, optional
            Time units for the data, by default None
        **kwargs : dict
            Additional keyword arguments passed to parent class
        """
        super().__init__(data_path, dataset_name, time_unit, **kwargs)
        self.station_info = None
        self.basin_station_mapping = None

    def set_data_source_describe(self):
        """Set data source description including stations directory."""
        # Get the base description from parent class
        data_source_description = super().set_data_source_describe()

        # Add stations-related paths
        data_root_dir = self.data_source_dir
        stations_dir = os.path.join(data_root_dir, "stations")
        station_ts_dirs = self._where_ts_dir(stations_dir)
        basin_station_info_dir = os.path.join(stations_dir, "basin_station_info")

        # Add stations paths to description
        data_source_description["STATIONS_DIR"] = stations_dir
        data_source_description["STATION_TS_DIRS"] = station_ts_dirs
        data_source_description["BASIN_STATION_INFO_DIR"] = basin_station_info_dir

        return data_source_description

    def read_station_info(self):
        """Read basic station information and basin-station mapping."""
        info_dir = self.data_source_description["BASIN_STATION_INFO_DIR"]

        # Read all_basin_station_mapping.csv
        mapping_file = os.path.join(info_dir, "all_basin_station_mapping.csv")
        if "s3://" in mapping_file:
            with conf.FS.open(mapping_file, mode="rb") as f:
                mapping_data = pd.read_csv(
                    f, dtype={"basin_id": str, "station_id": str}
                )
        else:
            mapping_data = pd.read_csv(
                mapping_file, dtype={"basin_id": str, "station_id": str}
            )

        # Read basin_summary.csv
        summary_file = os.path.join(info_dir, "basin_summary.csv")
        if "s3://" in summary_file:
            with conf.FS.open(summary_file, mode="rb") as f:
                summary_data = pd.read_csv(f, dtype={"basin_id": str})
        else:
            summary_data = pd.read_csv(summary_file, dtype={"basin_id": str})

        self.basin_station_mapping = mapping_data
        self.station_info = summary_data

        return mapping_data, summary_data

    def read_station_object_ids(self) -> np.array:
        """Get all station IDs."""
        if self.basin_station_mapping is None:
            self.read_station_info()
        return self.basin_station_mapping["station_id"].unique()

    def read_basin_stations(self, basin_id: str) -> pd.DataFrame:
        """Read detailed station information for a specific basin.

        Parameters
        ----------
        basin_id : str
            Basin ID

        Returns
        -------
        pd.DataFrame
            Station details for the basin
        """
        info_dir = self.data_source_description["BASIN_STATION_INFO_DIR"]
        station_file = os.path.join(info_dir, f"basin_{basin_id}_stations.csv")

        if "s3://" in station_file:
            with conf.FS.open(station_file, mode="rb") as f:
                station_data = pd.read_csv(f, dtype={"station_id": str})
        else:
            station_data = pd.read_csv(station_file, dtype={"station_id": str})

        return station_data

    def read_basin_adjacency(self, basin_id: str) -> pd.DataFrame:
        """Read adjacency matrix for a specific basin.

        Parameters
        ----------
        basin_id : str
            Basin ID

        Returns
        -------
        pd.DataFrame
            Adjacency matrix for stations in the basin
        """
        info_dir = self.data_source_description["BASIN_STATION_INFO_DIR"]
        adjacency_file = os.path.join(info_dir, f"adjacency_{basin_id}_True.csv")

        if "s3://" in adjacency_file:
            with conf.FS.open(adjacency_file, mode="rb") as f:
                adjacency_data = pd.read_csv(f, index_col=0)
        else:
            adjacency_data = pd.read_csv(adjacency_file, index_col=0)

        return adjacency_data

    def read_station_timeseries(
        self, station_ids=None, t_range_list: list = None, relevant_cols=None, **kwargs
    ) -> dict:
        """Read timeseries data for stations.

        Parameters
        ----------
        station_ids : list, optional
            List of station IDs, by default None
        t_range_list : list, optional
            Time range [start_time, end_time], by default None
        relevant_cols : list, optional
            List of relevant columns, by default None
        **kwargs : dict
            Additional keyword arguments

        Returns
        -------
        dict
            Dictionary containing data with different time scales
        """
        time_units = kwargs.get("time_units", ["1D"])
        start0101_freq = kwargs.get("start0101_freq", False)
        offset_to_utc = kwargs.get("offset_to_utc", self.offset_to_utc)

        results = {}

        for time_unit in time_units:
            # Remove the forced offset_to_utc=True for 3h data
            # The offset should be controlled by the parameter or instance setting
            # if time_unit == "3h":
            #     offset_to_utc = True

            # Get station data directory for this time unit
            station_ts_dir = self._get_station_ts_dir(
                self.data_source_description["STATION_TS_DIRS"], time_unit
            )

            if start0101_freq:
                t_range = generate_start0101_time_range(
                    start_time=t_range_list[0],
                    end_time=t_range_list[-1],
                    freq=time_unit,
                )
            else:
                t_range = pd.date_range(
                    start=t_range_list[0], end=t_range_list[-1], freq=time_unit
                )

            nt = len(t_range)
            x = np.full([len(station_ids), nt, len(relevant_cols)], np.nan)

            for k in tqdm(
                range(len(station_ids)), desc=f"Reading station data with {time_unit}"
            ):
                station_file = os.path.join(
                    station_ts_dir,
                    station_ids[k] + ".csv",
                )

                if "s3://" in station_file:
                    with conf.FS.open(station_file, mode="rb") as f:
                        station_data = pd.read_csv(f, engine="c")
                else:
                    station_data = pd.read_csv(station_file, engine="c")

                # Check for different time column names
                time_col = None
                for col in ["time", "TM", "Time", "DATE", "date"]:
                    if col in station_data.columns:
                        time_col = col
                        break

                if time_col is None:
                    print(f"Warning: No time column found in {station_file}")
                    continue

                date = pd.to_datetime(station_data[time_col]).values
                if offset_to_utc:
                    # For station data, we might need a different offset handling
                    # For now, use a default offset, but this could be customized
                    date = date - np.timedelta64(8, "h")  # Default Beijing time offset

                [_, ind1, ind2] = np.intersect1d(date, t_range, return_indices=True)

                for j in range(len(relevant_cols)):
                    if relevant_cols[j] in station_data.columns:
                        tmp_ = station_data[relevant_cols[j]].values
                        x[k, ind2, j] = tmp_[ind1]

            results[time_unit] = x

        return results

    def _get_station_ts_dir(self, station_ts_dirs, time_unit):
        """Get the station timeseries directory for a specific time unit."""
        station_ts_dir = next(
            dir_path
            for dir_path in station_ts_dirs
            if time_unit == dir_path.split(os.sep)[-1]
        )
        # For station data, we don't append version to directory path
        # The version is only used for caching and file naming
        return station_ts_dir

    def cache_station_timeseries_xrdataset(self, **kwargs):
        """Cache all station timeseries data in separate NetCDF files.

        Parameters
        ----------
        **kwargs : dict
            batchsize -- Number of stations to process per batch, by default 100
            time_units -- List of time units to process, by default None
            start0101_freq -- Whether to use start0101 frequency, by default False
        """
        batchsize = kwargs.get("batchsize", 100)
        time_units = kwargs.get("time_units", self.time_unit) or ["1D"]
        start0101_freq = kwargs.get("start0101_freq", False)
        offset_to_utc = kwargs.get("offset_to_utc", self.offset_to_utc)

        # Get all station IDs
        station_ids = self.read_station_object_ids()

        # Define generator for batching
        def station_data_generator(stations, batch_size):
            for i in range(0, len(stations), batch_size):
                yield stations[i : i + batch_size]

        for time_unit in time_units:
            if self.trange4cache is None:
                if time_unit != "3h":
                    self.trange4cache = ["1960-01-01", "2024-12-31"]
                else:
                    self.trange4cache = ["1960-01-01 02", "2024-12-31 23"]

            # Generate time range
            if start0101_freq:
                times = (
                    generate_start0101_time_range(
                        start_time=self.trange4cache[0],
                        end_time=self.trange4cache[-1],
                        freq=time_unit,
                    )
                    .strftime("%Y-%m-%d %H:%M:%S")
                    .tolist()
                )
            else:
                times = (
                    pd.date_range(
                        start=self.trange4cache[0],
                        end=self.trange4cache[-1],
                        freq=time_unit,
                    )
                    .strftime("%Y-%m-%d %H:%M:%S")
                    .tolist()
                )

            # Get sample station data to determine variables
            sample_station_file = os.path.join(
                self._get_station_ts_dir(
                    self.data_source_description["STATION_TS_DIRS"], time_unit
                ),
                station_ids[0] + ".csv",
            )

            if "s3://" in sample_station_file:
                with conf.FS.open(sample_station_file, mode="rb") as f:
                    sample_data = pd.read_csv(f, engine="c")
            else:
                sample_data = pd.read_csv(sample_station_file, engine="c")

            # Find time column and exclude it from variables
            time_col = None
            for col in ["time", "TM", "Time", "DATE", "date"]:
                if col in sample_data.columns:
                    time_col = col
                    break

            variables = [col for col in sample_data.columns if col != time_col]

            # Get units info if available
            units_info = {var: "unknown" for var in variables}

            for station_batch in station_data_generator(station_ids, batchsize):
                data = self.read_station_timeseries(
                    station_ids=station_batch,
                    t_range_list=self.trange4cache,
                    relevant_cols=variables,
                    time_units=[time_unit],
                    start0101_freq=start0101_freq,
                    offset_to_utc=offset_to_utc,
                )

                dataset = xr.Dataset(
                    data_vars={
                        variables[i]: (
                            ["station", "time"],
                            data[time_unit][:, :, i],
                            {"units": units_info.get(variables[i], "unknown")},
                        )
                        for i in range(len(variables))
                    },
                    coords={
                        "station": station_batch,
                        "time": pd.to_datetime(times),
                    },
                )

                # Save dataset
                prefix_ = self._get_station_file_prefix_(
                    self.dataset_name, self.version
                )
                batch_file_path = os.path.join(
                    CACHE_DIR,
                    f"{prefix_}stations_{time_unit}_batch_{station_batch[0]}_{station_batch[-1]}.nc",
                )
                dataset.to_netcdf(batch_file_path)

                # Clean up memory
                del dataset
                del data

    def cache_station_info_xrdataset(self):
        """Cache station information and basin-station relationships."""
        # Read station info
        mapping_data, summary_data = self.read_station_info()

        # Cache basin-station mapping
        mapping_ds = xr.Dataset(
            data_vars={
                col: (["mapping_id"], mapping_data[col].values)
                for col in mapping_data.columns
            },
            coords={"mapping_id": range(len(mapping_data))},
        )

        # Cache basin summary
        summary_ds = xr.Dataset(
            data_vars={
                col: (["basin"], summary_data[col].values)
                for col in summary_data.columns
                if col != "basin_id"
            },
            coords={"basin": summary_data["basin_id"].values},
        )

        # Save datasets
        dataset_name = self.dataset_name
        prefix_ = "" if dataset_name is None else dataset_name + "_"

        mapping_ds.to_netcdf(
            os.path.join(CACHE_DIR, f"{prefix_}basin_station_mapping.nc")
        )
        summary_ds.to_netcdf(os.path.join(CACHE_DIR, f"{prefix_}basin_summary.nc"))

    def read_station_ts_xrdataset(
        self,
        station_id_lst: list = None,
        t_range: list = None,
        var_lst: list = None,
        **kwargs,
    ) -> dict:
        """Read station timeseries data from cached NetCDF files.

        Parameters
        ----------
        station_id_lst : list, optional
            List of station IDs to select, by default None
        t_range : list, optional
            Time range [start_time, end_time], by default None
        var_lst : list, optional
            List of variables to select, by default None
        **kwargs : dict
            Additional arguments

        Returns
        -------
        dict
            Dictionary with time units as keys and xarray.Dataset as values
        """
        time_units = kwargs.get("time_units", self.time_unit)
        # if var_lst is None:
        #     return None

        datasets_by_time_unit = {}
        prefix_ = self._get_station_file_prefix_(self.dataset_name, self.version)

        for time_unit in time_units:
            batch_files = self._get_station_batch_files(prefix_, time_unit)

            if not batch_files:
                # Cache data if not found
                self.cache_station_timeseries_xrdataset(**kwargs)
                batch_files = self._get_station_batch_files(prefix_, time_unit)

            selected_datasets = []

            for batch_file in batch_files:
                ds = xr.open_dataset(batch_file)

                if any(var not in ds.variables for var in var_lst):
                    raise ValueError(f"var_lst must all be in {ds.data_vars}")

                if valid_station_ids := [
                    sid for sid in station_id_lst if sid in ds["station"].values
                ]:
                    ds_selected = ds[var_lst].sel(
                        station=valid_station_ids, time=slice(t_range[0], t_range[1])
                    )
                    selected_datasets.append(ds_selected)

                ds.close()

            if selected_datasets:
                datasets_by_time_unit[time_unit] = xr.concat(
                    selected_datasets, dim="station"
                ).sortby("station")
            else:
                datasets_by_time_unit[time_unit] = xr.Dataset()

        return datasets_by_time_unit

    def read_station_info_xrdataset(self, **kwargs):
        """Read station information from cached NetCDF files.

        Returns
        -------
        tuple
            (basin_station_mapping_dataset, basin_summary_dataset)
        """
        dataset_name = self.dataset_name
        prefix_ = "" if dataset_name is None else dataset_name + "_"

        try:
            mapping_ds = xr.open_dataset(
                os.path.join(CACHE_DIR, f"{prefix_}basin_station_mapping.nc")
            )
            summary_ds = xr.open_dataset(
                os.path.join(CACHE_DIR, f"{prefix_}basin_summary.nc")
            )
        except FileNotFoundError:
            self.cache_station_info_xrdataset()
            mapping_ds = xr.open_dataset(
                os.path.join(CACHE_DIR, f"{prefix_}basin_station_mapping.nc")
            )
            summary_ds = xr.open_dataset(
                os.path.join(CACHE_DIR, f"{prefix_}basin_summary.nc")
            )

        return mapping_ds, summary_ds

    def cache_adjacency_xrdataset(self):
        """Cache adjacency matrices for all basins as NetCDF files."""
        # Get all basin IDs
        if self.basin_station_mapping is None:
            self.read_station_info()

        basin_ids = self.basin_station_mapping["basin_id"].unique()

        # Read all adjacency matrices
        adjacency_datasets = {}
        for basin_id in basin_ids:
            try:
                adjacency_data = self.read_basin_adjacency(basin_id)
                # Convert to xarray Dataset, keeping string types
                adjacency_ds = xr.Dataset(
                    data_vars={
                        col: (["ID"], adjacency_data[col].values.astype(str))
                        for col in adjacency_data.columns
                    },
                    coords={"ID": adjacency_data.index.values.astype(str)},
                )
                adjacency_datasets[basin_id] = adjacency_ds
            except FileNotFoundError:
                # Skip basins without adjacency files
                continue

        # Save individual adjacency matrices for each basin
        dataset_name = self.dataset_name
        prefix_ = "" if dataset_name is None else dataset_name + "_"

        for basin_id, adj_ds in adjacency_datasets.items():
            adj_ds.to_netcdf(
                os.path.join(CACHE_DIR, f"{prefix_}adjacency_{basin_id}.nc")
            )

    def read_adjacency_xrdataset(self, basin_id: str) -> xr.Dataset:
        """Read cached adjacency matrix for a specific basin from NetCDF file.

        Parameters
        ----------
        basin_id : str
            Basin ID

        Returns
        -------
        xr.Dataset
            Adjacency matrix dataset for the basin
        """
        dataset_name = self.dataset_name
        prefix_ = "" if dataset_name is None else dataset_name + "_"

        adjacency_file = os.path.join(CACHE_DIR, f"{prefix_}adjacency_{basin_id}.nc")

        try:
            adjacency_ds = xr.open_dataset(adjacency_file)
        except FileNotFoundError:
            # Cache the adjacency data if not found
            self.cache_adjacency_xrdataset()
            adjacency_ds = xr.open_dataset(adjacency_file)

        return adjacency_ds

    def _get_station_file_prefix_(self, dataset_name, version):
        """Get file prefix for station data files."""
        prefix_ = "" if dataset_name is None else dataset_name + "_"
        prefix_ = prefix_ + f"{version}_" if version is not None else prefix_
        return prefix_

    def _get_station_batch_files(self, prefix_, time_unit):
        """Get station batch files for a specific time unit."""
        return [
            os.path.join(CACHE_DIR, f)
            for f in os.listdir(CACHE_DIR)
            if re.match(
                rf"^{prefix_}stations_{time_unit}_batch_[A-Za-z0-9_]+_[A-Za-z0-9_]+\.nc$",
                f,
            )
        ]

    def cache_all_station_data(self, **kwargs):
        """Cache all station-related data including timeseries, info and adjacency."""
        self.cache_station_timeseries_xrdataset(**kwargs)
        self.cache_station_info_xrdataset()
        self.cache_adjacency_xrdataset()

    def get_stations_by_basin(self, basin_id: str) -> list:
        """Get all station IDs for a specific basin.

        Parameters
        ----------
        basin_id : str
            Basin ID

        Returns
        -------
        list
            List of station IDs in the basin
        """
        if self.basin_station_mapping is None:
            self.read_station_info()

        return (
            self.basin_station_mapping[
                self.basin_station_mapping["basin_id"] == basin_id
            ]["station_id"]
            .unique()
            .tolist()
        )



"""
Author: Wenyu Ouyang
Date: 2021-12-05 11:21:58
LastEditTime: 2025-11-03 17:29:39
LastEditors: Wenyu Ouyang
Description: Data source class for Gages
FilePath: \hydrodatasource\hydrodatasource\reader\gages.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

import collections
import logging
import os

import numpy as np
import pandas as pd
import xarray as xr
from datetime import datetime, timedelta
from typing import Tuple, Dict, Union
import pytz
from pandas.core.dtypes.common import is_string_dtype, is_numeric_dtype
from tqdm import tqdm
from hydroutils import hydro_logger, hydro_file, hydro_stat
from hydrodatasource.reader.data_source import HydroData


class Gages(HydroData):
    def __init__(self, data_path, dataset_name):
        super().__init__(data_path, dataset_name)
        self.data_source_description = self.set_data_source_describe()
        self.gages_sites = self.read_site_info()

    def get_name(self):
        return "GAGES"

    def get_constant_cols(self) -> np.array:
        """all readable attrs in GAGES-II"""
        dir_gage_attr = self.data_source_description["GAGES_ATTR_DIR"]
        var_desc_file = os.path.join(dir_gage_attr, "variable_descriptions.txt")
        var_desc = pd.read_csv(var_desc_file)
        return var_desc["VARIABLE_NAME"].values

    def get_relevant_cols(self):
        return np.array(["dayl", "prcp", "srad", "swe", "tmax", "tmin", "vp"])

    def get_target_cols(self):
        return np.array(["usgsFlow"])

    def get_other_cols(self) -> dict:
        return {
            "FDC": {"time_range": ["1980-01-01", "2000-01-01"], "quantile_num": 100}
        }

    def set_data_source_describe(self):
        gages_db = self.data_source_dir
        # region shapefiles
        gage_region_dir = os.path.join(
            gages_db,
            "boundaries_shapefiles_by_aggeco",
            "boundaries-shapefiles-by-aggeco",
        )
        gages_regions = [
            "bas_ref_all",
            "bas_nonref_CntlPlains",
            "bas_nonref_EastHghlnds",
            "bas_nonref_MxWdShld",
            "bas_nonref_NorthEast",
            "bas_nonref_SECstPlain",
            "bas_nonref_SEPlains",
            "bas_nonref_WestMnts",
            "bas_nonref_WestPlains",
            "bas_nonref_WestXeric",
        ]
        # point shapefile
        gagesii_points_file = os.path.join(
            gages_db, "gagesII_9322_point_shapefile", "gagesII_9322_sept30_2011.shp"
        )

        # config of flow data
        flow_dir = os.path.join(gages_db, "gages_streamflow", "gages_streamflow")
        # forcing
        forcing_dir = os.path.join(gages_db, "basin_mean_forcing", "basin_mean_forcing")
        forcing_types = ["daymet"]
        # attr
        attr_dir = os.path.join(
            gages_db, "basinchar_and_report_sept_2011", "spreadsheets-in-csv-format"
        )
        gauge_id_file = os.path.join(attr_dir, "conterm_basinid.txt")

        download_url_lst = [
            "https://water.usgs.gov/GIS/dsdl/basinchar_and_report_sept_2011.zip",
            "https://water.usgs.gov/GIS/dsdl/gagesII_9322_point_shapefile.zip",
            "https://water.usgs.gov/GIS/dsdl/boundaries_shapefiles_by_aggeco.zip",
            "https://www.sciencebase.gov/catalog/file/get/59692a64e4b0d1f9f05fbd39",
        ]
        usgs_streamflow_url = "https://waterdata.usgs.gov/nwis/dv?cb_00060=on&format=rdb&site_no={}&referred_module=sw&period=&begin_date={}-{}-{}&end_date={}-{}-{}"
        # GAGES-II time series data_source dir
        gagests_dir = os.path.join(gages_db, "59692a64e4b0d1f9f05f")
        population_file = os.path.join(
            gagests_dir,
            "Dataset8_Population-Housing",
            "Dataset8_Population-Housing",
            "PopulationHousing.txt",
        )
        wateruse_file = os.path.join(
            gagests_dir,
            "Dataset10_WaterUse",
            "Dataset10_WaterUse",
            "WaterUse_1985-2010.txt",
        )
        return collections.OrderedDict(
            GAGES_DIR=gages_db,
            GAGES_FLOW_DIR=flow_dir,
            GAGES_FORCING_DIR=forcing_dir,
            GAGES_FORCING_TYPE=forcing_types,
            GAGES_ATTR_DIR=attr_dir,
            GAGES_GAUGE_FILE=gauge_id_file,
            GAGES_DOWNLOAD_URL_LST=download_url_lst,
            GAGES_REGIONS_SHP_DIR=gage_region_dir,
            GAGES_REGION_LIST=gages_regions,
            GAGES_POINT_SHP_FILE=gagesii_points_file,
            GAGES_POPULATION_FILE=population_file,
            GAGES_WATERUSE_FILE=wateruse_file,
            USGS_FLOW_URL=usgs_streamflow_url,
        )

    def read_other_cols(self, object_ids=None, other_cols=None, **kwargs) -> dict:
        # TODO: not finish
        out_dict = {}
        for key, value in other_cols.items():
            if key != "FDC":
                raise NotImplementedError("No this item yet!!")
            assert "time_range" in value.keys()
            if "quantile_num" in value.keys():
                quantile_num = value["quantile_num"]
                out = hydro_stat.cal_fdc(
                    self.read_target_cols(object_ids, value["time_range"], "usgsFlow"),
                    quantile_num=quantile_num,
                )
            else:
                out = hydro_stat.cal_fdc(
                    self.read_target_cols(object_ids, value["time_range"], "usgsFlow")
                )
            out_dict[key] = out
        return out_dict

    def read_attr_all(self, gages_ids: Union[list, np.ndarray]):
        """
        read all attr data for some sites in GAGES-II
        TODO: now it is not same as functions in CAMELS where read_attr_all has no "gages_ids" parameter

        Parameters
        ----------
        gages_ids : Union[list, np.ndarray]
            gages sites' ids

        Returns
        -------
        ndarray
            all attr data for gages_ids
        """
        dir_gage_attr = self.data_source_description["GAGES_ATTR_DIR"]
        f_dict = {}  # factorize dict
        # each key-value pair for atts in a file (list）
        var_dict = {}
        # all attrs
        var_lst = []
        out_lst = []
        key_lst = self._get_attr_col_names(dir_gage_attr)
        for key in key_lst:
            # in "spreadsheets-in-csv-format" directory, the name of "flow_record" file is conterm_flowrec.txt
            if key == "flow_record":
                key = "flowrec"
            data_file = os.path.join(dir_gage_attr, "conterm_" + key + ".txt")
            # remove some unused atttrs in bas_classif
            if key == "bas_classif":
                # https://stackoverflow.com/questions/22216076/unicodedecodeerror-utf8-codec-cant-decode-byte-0xa5-in-position-0-invalid-s
                data_temp = pd.read_csv(
                    data_file,
                    sep=",",
                    dtype={"STAID": str},
                    usecols=range(4),
                    encoding="unicode_escape",
                )
            else:
                data_temp = pd.read_csv(data_file, sep=",", dtype={"STAID": str})
            if key == "flowrec":
                # remove final column which is nan
                data_temp = data_temp.iloc[:, range(data_temp.shape[1] - 1)]
            # all attrs in files
            var_lst_temp = list(data_temp.columns[1:])
            var_dict[key] = var_lst_temp
            var_lst.extend(var_lst_temp)
            k = 0
            n_gage = len(gages_ids)
            out_temp = np.full(
                [n_gage, len(var_lst_temp)], np.nan
            )  # 1d:sites，2d: attrs in current data_file
            # sites intersection，ind2 is the index of sites in conterm_ files，set them in out_temp
            range1 = gages_ids
            range2 = data_temp.iloc[:, 0].astype(str).tolist()
            assert all(x < y for x, y in zip(range2, range2[1:]))
            # Notice the sequence of station ids ! Some id_lst_all are not sorted, so don't use np.intersect1d
            ind2 = [range2.index(tmp) for tmp in range1]
            for field in var_lst_temp:
                if is_string_dtype(data_temp[field]):  # str vars -> categorical vars
                    value, ref = pd.factorize(data_temp.loc[ind2, field], sort=True)
                    out_temp[:, k] = value
                    f_dict[field] = ref.tolist()
                elif is_numeric_dtype(data_temp[field]):
                    out_temp[:, k] = data_temp.loc[ind2, field].values
                k = k + 1
            out_lst.append(out_temp)
        out = np.concatenate(out_lst, 1)
        return out, var_lst, var_dict, f_dict

    def read_constant_cols(
        self, object_ids=None, constant_cols: list = None, **kwargs
    ) -> np.array:
        """
        read some attrs of some sites

        Parameters
        ----------
        object_ids : [type], optional
            sites_ids, by default None
        constant_cols : list, optional
            attrs' names, by default None

        Returns
        -------
        np.array
            attr data for object_ids
        """
        # assert all(x < y for x, y in zip(object_ids, object_ids[1:]))
        attr_all, var_lst_all, var_dict, f_dict = self.read_attr_all(object_ids)
        ind_var = [var_lst_all.index(var) for var in constant_cols]
        return attr_all[:, ind_var]

    def read_attr_origin(self, gages_ids, attr_lst) -> np.ndarray:
        """
        this function read the attrs data in GAGES-II but not transform them to int when they are str

        Parameters
        ----------
        gages_ids : [type]
            [description]
        attr_lst : [type]
            [description]

        Returns
        -------
        np.ndarray
            the first dim is types of attrs, and the second one is sites
        """
        dir_gage_attr = self.data_source_description["GAGES_ATTR_DIR"]
        key_lst = self._get_attr_col_names(dir_gage_attr)
        out_lst = []
        out_lst.extend([] for _ in range(len(attr_lst)))
        range1 = gages_ids
        gage_id_file = self.data_source_description["GAGES_GAUGE_FILE"]
        data_all = pd.read_csv(gage_id_file, sep=",", dtype={0: str})
        range2 = data_all["STAID"].values.tolist()
        assert all(x < y for x, y in zip(range2, range2[1:]))
        # Notice the sequence of station ids ! Some id_lst_all are not sorted, so don't use np.intersect1d
        ind2 = [range2.index(tmp) for tmp in range1]

        for key in key_lst:
            # in "spreadsheets-in-csv-format" directory, the name of "flow_record" file is conterm_flowrec.txt
            if key == "flow_record":
                key = "flowrec"
            data_file = os.path.join(dir_gage_attr, "conterm_" + key + ".txt")
            if key == "bas_classif":
                data_temp = pd.read_csv(
                    data_file,
                    sep=",",
                    dtype={
                        "STAID": str,
                        "WR_REPORT_REMARKS": str,
                        "ADR_CITATION": str,
                        "SCREENING_COMMENTS": str,
                    },
                    engine="python",
                    encoding="unicode_escape",
                )
            elif key == "bound_qa":
                # "DRAIN_SQKM" already exists
                data_temp = pd.read_csv(
                    data_file,
                    sep=",",
                    dtype={"STAID": str},
                    usecols=[
                        "STAID",
                        "BASIN_BOUNDARY_CONFIDENCE",
                        "NWIS_DRAIN_SQKM",
                        "PCT_DIFF_NWIS",
                        "HUC10_CHECK",
                    ],
                )
            else:
                data_temp = pd.read_csv(data_file, sep=",", dtype={"STAID": str})
            if key == "flowrec":
                data_temp = data_temp.iloc[:, range(data_temp.shape[1] - 1)]
            var_lst_temp = list(data_temp.columns[1:])
            do_exist, idx_lst = is_any_elem_in_a_lst(
                attr_lst, var_lst_temp, return_index=True
            )
            if do_exist:
                for idx in idx_lst:
                    idx_in_var = (
                        var_lst_temp.index(attr_lst[idx]) + 1
                    )  # +1 because the first col of data_temp is ID
                    out_lst[idx] = data_temp.iloc[ind2, idx_in_var].values
            else:
                continue
        return np.array(out_lst)

    def _get_attr_col_names(self, dir_gage_attr):
        var_des = pd.read_csv(
            os.path.join(dir_gage_attr, "variable_descriptions.txt"), sep=","
        )
        var_des_map_values = var_des["VARIABLE_TYPE"].tolist()
        for i in range(len(var_des)):
            var_des_map_values[i] = var_des_map_values[i].lower()
        result = list(set(var_des_map_values))
        result.sort(key=var_des_map_values.index)
        result.remove("x_region_names")
        return result

    def read_forcing_gage(self, usgs_id, var_lst, t_range_list, forcing_type="daymet"):
        gage_dict = self.gages_sites
        ind = np.argwhere(gage_dict["STAID"] == usgs_id)[0][0]
        huc = gage_dict["HUC02"][ind]

        data_folder = os.path.join(
            self.data_source_description["GAGES_FORCING_DIR"], forcing_type
        )
        # original daymet file not for leap year, there is no data in 12.31 in leap year,
        # so files which have been interpolated for nan value have name "_leap"
        data_file = os.path.join(
            data_folder, huc, f"{usgs_id}_lump_{forcing_type}_forcing_leap.txt"
        )
        data_temp = pd.read_csv(data_file, sep=r"\s+", header=None, skiprows=1)

        df_date = data_temp[[0, 1, 2]]
        df_date.columns = ["year", "month", "day"]
        date = pd.to_datetime(df_date).values.astype("datetime64[D]")
        nf = len(var_lst)
        assert all(x < y for x, y in zip(date, date[1:]))
        [c, ind1, ind2] = np.intersect1d(date, t_range_list, return_indices=True)
        assert date[0] <= t_range_list[0] and date[-1] >= t_range_list[-1]
        nt = t_range_list.size
        out = np.empty([nt, nf])
        var_lst_in_file = [
            "dayl(s)",
            "prcp(mm/day)",
            "srad(W/m2)",
            "swe(mm)",
            "tmax(C)",
            "tmin(C)",
            "vp(Pa)",
        ]
        for k in range(nf):
            # assume all files are of same columns. May check later.
            ind = [
                i
                for i in range(len(var_lst_in_file))
                if var_lst[k] in var_lst_in_file[i]
            ][0]
            out[ind2, k] = data_temp[ind + 4].values[ind1]
        return out

    def read_relevant_cols(
        self, object_ids=None, t_range_list=None, var_lst=None, **kwargs
    ) -> np.array:
        assert all(x < y for x, y in zip(object_ids, object_ids[1:]))
        assert all(x < y for x, y in zip(t_range_list, t_range_list[1:]))
        t_lst = hydro_utils.t_range_days(t_range_list)
        nt = t_lst.shape[0]
        x = np.empty([len(object_ids), nt, len(var_lst)])
        for k in tqdm(range(len(object_ids)), desc="reading GAGES forcing data"):
            data = self.read_forcing_gage(
                object_ids[k],
                var_lst,
                t_lst,
                forcing_type=self.data_source_description["GAGES_FORCING_TYPE"][0],
            )
            x[k, :, :] = data
        return x

    def read_target_cols(
        self, usgs_id_lst=None, t_range_list=None, target_cols=None, **kwargs
    ) -> np.array:
        """
        Read USGS daily average streamflow data according to id and time

        Parameters
        ----------
        usgs_id_lst
            site information
        t_range_list
            must be time range for downloaded data
        target_cols

        kwargs
            optional

        Returns
        -------
        np.array
            streamflow data, 1d-axis: gages, 2d-axis: day, 3d-axis: streamflow
        """
        t_lst = hydro_utils.t_range_days(t_range_list)
        nt = t_lst.shape[0]
        y = np.empty([len(usgs_id_lst), nt, 1])
        for k in tqdm(range(len(usgs_id_lst)), desc="Read GAGES streamflow data"):
            data_obs = self.read_usgs_gage(usgs_id_lst[k], t_lst)
            y[k, :, 0] = data_obs
        return y

    def read_usgs_gage(self, usgs_id, t_lst):
        """
        read data for one gage

        Parameters
        ----------
        usgs_id : [type]
            [description]
        t_lst : [type]
            [description]

        Returns
        -------
        [type]
            [description]
        """
        dir_gage_flow = self.data_source_description["GAGES_FLOW_DIR"]
        gage_id_df = pd.DataFrame(self.gages_sites)
        huc = gage_id_df[gage_id_df["STAID"] == usgs_id]["HUC02"].values[0]
        usgs_file = os.path.join(dir_gage_flow, str(huc), usgs_id + ".txt")
        # ignore the comment lines and the first non-value row
        df_flow = pd.read_csv(
            usgs_file, comment="#", sep="\t", dtype={"site_no": str}
        ).iloc[1:, :]
        # reset the index, start from 0
        df_flow = df_flow.reset_index(drop=True)
        # change the original column names
        columns_names = df_flow.columns.tolist()
        columns_flow = []
        columns_flow.extend(
            column_name
            for column_name in columns_names
            if "_00060_00003" in column_name and "_00060_00003_cd" not in column_name
        )
        columns_flow_cd = [
            column_name
            for column_name in columns_names
            if "_00060_00003_cd" in column_name
        ]
        if len(columns_flow) > 1:
            self._chose_flow_col(df_flow, t_lst, columns_flow, columns_flow_cd)
        else:
            for column_name in columns_names:
                if (
                    "_00060_00003" in column_name
                    and "_00060_00003_cd" not in column_name
                ):
                    df_flow.rename(columns={column_name: "flow"}, inplace=True)
                    break
            for column_name in columns_names:
                if "_00060_00003_cd" in column_name:
                    df_flow.rename(columns={column_name: "mode"}, inplace=True)
                    break

        columns = ["agency_cd", "site_no", "datetime", "flow", "mode"]
        if df_flow.empty:
            df_flow = pd.DataFrame(columns=columns)
        if "flow" not in df_flow.columns.intersection(columns):
            data_temp = df_flow.loc[:, df_flow.columns.intersection(columns)]
            # add nan column to data_temp
            data_temp = pd.concat([data_temp, pd.DataFrame(columns=["flow", "mode"])])
        else:
            data_temp = df_flow.loc[:, columns]
        self._set_some_value_to_nan(data_temp, "flow")
        # set negative value -- nan
        obs = data_temp["flow"].astype("float").values
        obs[obs < 0] = np.nan
        # time range intersection. set points without data nan values
        nt = len(t_lst)
        out = np.full([nt], np.nan)
        # date in df is str，so transform them to datetime
        df_date = data_temp["datetime"]
        date = pd.to_datetime(df_date).values.astype("datetime64[D]")
        c, ind1, ind2 = np.intersect1d(date, t_lst, return_indices=True)
        out[ind2] = obs[ind1]
        return out

    def _chose_flow_col(self, df_flow, t_lst, columns_flow, columns_flow_cd):
        logging.debug("there are some columns for flow, choose one\n")
        df_date_temp = df_flow["datetime"]
        date_temp = pd.to_datetime(df_date_temp).values.astype("datetime64[D]")
        c_temp, ind1_temp, ind2_temp = np.intersect1d(
            date_temp, t_lst, return_indices=True
        )
        num_nan_lst = []
        for item in columns_flow:
            out_temp = np.full([len(t_lst)], np.nan)

            self._set_some_value_to_nan(df_flow, item)
            df_flow_temp = df_flow[item].copy()
            out_temp[ind2_temp] = df_flow_temp[ind1_temp]
            num_nan = np.isnan(out_temp).sum()
            num_nan_lst.append(num_nan)
        num_nan_np = np.array(num_nan_lst)
        index_flow_num = np.argmin(num_nan_np)
        df_flow.rename(columns={columns_flow[index_flow_num]: "flow"}, inplace=True)
        df_flow.rename(columns={columns_flow_cd[index_flow_num]: "mode"}, inplace=True)

    def _set_some_value_to_nan(self, arg0, arg1):
        arg0.loc[arg0[arg1] == "Ice", arg1] = np.nan
        arg0.loc[arg0[arg1] == "Ssn", arg1] = np.nan
        arg0.loc[arg0[arg1] == "Tst", arg1] = np.nan
        arg0.loc[arg0[arg1] == "Eqp", arg1] = np.nan
        arg0.loc[arg0[arg1] == "Rat", arg1] = np.nan
        arg0.loc[arg0[arg1] == "Dis", arg1] = np.nan
        arg0.loc[arg0[arg1] == "Bkw", arg1] = np.nan
        arg0.loc[arg0[arg1] == "***", arg1] = np.nan
        arg0.loc[arg0[arg1] == "Mnt", arg1] = np.nan
        arg0.loc[arg0[arg1] == "ZFL", arg1] = np.nan

    def read_object_ids(self, object_params=None) -> np.array:
        return self.gages_sites["STAID"]

    def read_basin_area(self, object_ids) -> np.array:
        return self.read_constant_cols(object_ids, ["DRAIN_SQKM"], is_return_dict=False)

    def read_mean_prep(self, object_ids) -> np.array:
        mean_prep = self.read_constant_cols(
            object_ids, ["PPTAVG_BASIN"], is_return_dict=False
        )
        mean_prep = mean_prep / 365 * 10
        return mean_prep

    def download_data_source(self):
        print("Please download data manually!")
        if not os.path.isdir(self.data_source_description["GAGES_DIR"]):
            os.makedirs(self.data_source_description["GAGES_DIR"])
        zip_files = [
            "59692a64e4b0d1f9f05fbd39",
            "basin_mean_forcing.zip",
            "basinchar_and_report_sept_2011.zip",
            "boundaries_shapefiles_by_aggeco.zip",
            "gages_streamflow.zip",
            "gagesII_9322_point_shapefile.zip",
        ]
        download_zip_files = [
            os.path.join(self.data_source_description["GAGES_DIR"], zip_file)
            for zip_file in zip_files
        ]
        for download_zip_file in download_zip_files:
            if not os.path.isfile(download_zip_file):
                raise RuntimeError(
                    download_zip_file + " not found! Please download the data"
                )
        unzip_dirs = [
            os.path.join(self.data_source_description["GAGES_DIR"], zip_file[:-4])
            for zip_file in zip_files
        ]
        for i in range(len(unzip_dirs)):
            if not os.path.isdir(unzip_dirs[i]):
                print("unzip directory:" + unzip_dirs[i])
                unzip_nested_zip(download_zip_files[i], unzip_dirs[i])
            else:
                print("unzip directory -- " + unzip_dirs[i] + " has existed")

    def read_site_info(self):
        gage_id_file = self.data_source_description["GAGES_GAUGE_FILE"]
        data_all = pd.read_csv(gage_id_file, sep=",", dtype={0: str})
        gage_fld_lst = data_all.columns.values
        out = {}
        df_id_region = data_all.iloc[:, 0].values
        assert all(x < y for x, y in zip(df_id_region, df_id_region[1:]))
        for s in gage_fld_lst:
            if s is gage_fld_lst[1]:
                out[s] = data_all[s].values.tolist()
            else:
                out[s] = data_all[s].values
        return pd.DataFrame(out)

    def prepare_usgs_data(self):
        hydro_logger.info(
            "NOT all data_source could be downloaded from website directly!"
        )
        data_source_description = self.data_source_description
        # download zip files
        [
            hydro_file.download_one_zip(attr_url, data_source_description["GAGES_DIR"])
            for attr_url in data_source_description["GAGES_DOWNLOAD_URL_LST"]
        ]
        # download streamflow data from USGS website
        dir_gage_flow = data_source_description["GAGES_FLOW_DIR"]
        streamflow_url = data_source_description["USGS_FLOW_URL"]
        # TODO: now a hard code for t_download_range
        t_download_range = ["1980-01-01", "2025-01-01"]
        if not os.path.isdir(dir_gage_flow):
            os.makedirs(dir_gage_flow)
        dir_list = os.listdir(dir_gage_flow)
        # if no streamflow data for the usgs_id_lst, then download them from the USGS website
        data_all = pd.read_csv(
            data_source_description["GAGES_GAUGE_FILE"], sep=",", dtype={0: str}
        )
        usgs_id_lst = data_all.iloc[:, 0].values.tolist()
        gage_fld_lst = data_all.columns.values
        for ind in range(len(usgs_id_lst)):  # different hucs different directories
            huc_02 = data_all[gage_fld_lst[3]][ind]
            dir_huc_02 = str(huc_02)
            if dir_huc_02 not in dir_list:
                dir_huc_02 = os.path.join(dir_gage_flow, str(huc_02))
                os.mkdir(dir_huc_02)
                dir_list = os.listdir(dir_gage_flow)
            dir_huc_02 = os.path.join(dir_gage_flow, str(huc_02))
            file_list = os.listdir(dir_huc_02)
            file_usgs_id = f"{str(usgs_id_lst[ind])}.txt"
            if file_usgs_id not in file_list:
                # download data and save as txt file
                start_time_str = datetime.strptime(t_download_range[0], "%Y-%m-%d")
                end_time_str = datetime.strptime(
                    t_download_range[1], "%Y-%m-%d"
                ) - timedelta(days=1)
                url = streamflow_url.format(
                    usgs_id_lst[ind],
                    start_time_str.year,
                    start_time_str.month,
                    start_time_str.day,
                    end_time_str.year,
                    end_time_str.month,
                    end_time_str.day,
                )

                # save in its HUC02 dir
                temp_file = os.path.join(dir_huc_02, f"{str(usgs_id_lst[ind])}.txt")
                hydro_file.download_small_file(url, temp_file)
                print("successfully download " + temp_file + " streamflow data!")

    def cache_attributes_xrdataset(self):
        """Convert all the attributes to a single dataset

        Returns
        -------
        None
        """
        # NOTICE: although it seems that we don't use pint_xarray, we have to import this package
        import pint_xarray  # noqa: F401
        from hydrodatasource.configs import config as conf

        # 1. Get all site IDs
        object_ids = self.read_object_ids()

        # 2. Read all attributes
        attr_data, var_lst, _, f_dict = self.read_attr_all(object_ids)

        # Handle duplicate columns
        unique_vars = []
        unique_indices = []
        seen_vars = set()
        for i, var in enumerate(var_lst):
            if var not in seen_vars:
                unique_vars.append(var)
                unique_indices.append(i)
                seen_vars.add(var)

        attr_data_unique = attr_data[:, unique_indices]
        var_lst_unique = unique_vars

        # 3. Create a pandas DataFrame
        df_attr = pd.DataFrame(
            attr_data_unique, index=object_ids, columns=var_lst_unique
        )

        # Get units from variable_descriptions.txt
        dir_gage_attr = self.data_source_description["GAGES_ATTR_DIR"]
        var_desc_file = os.path.join(dir_gage_attr, "variable_descriptions.txt")
        try:
            var_desc = pd.read_csv(var_desc_file)
            units_dict = pd.Series(
                var_desc["UNITS (numeric values)"].values, index=var_desc.VARIABLE_NAME
            ).to_dict()
        except (FileNotFoundError, KeyError):
            units_dict = {}

        # 4. Create an xarray.Dataset
        ds = xr.Dataset()
        for column in df_attr.columns:
            attrs = {"units": units_dict.get(column, "unknown")}
            if column in f_dict:
                attrs["category_mapping"] = str(dict(enumerate(f_dict[column])))

            data_array = xr.DataArray(
                data=df_attr[column].values,
                dims=["basin"],
                coords={"basin": df_attr.index.values.astype(str)},
                attrs=attrs,
                name=column,
            )
            ds[column] = data_array

        # 5. Save the Dataset
        dataset_name = self.dataset_name
        prefix_ = "" if dataset_name is None else dataset_name + "_"
        ds.to_netcdf(os.path.join(conf.CACHE_DIR, f"{prefix_}attributes.nc"))

    def cache_timeseries_xrdataset(self, trange4cache=None, **kwargs):
        """Save all timeseries data in separate NetCDF files for each time unit.

        Parameters
        ----------
        trange4cache : list, optional
            Time range for caching data, by default ["1980-01-01", "2023-12-31"]
        kwargs : dict, optional
            batchsize -- Number of basins to process per batch, by default 100
            time_units -- List of time units to process, by default None
            start0101_freq -- for freq setting, if the start date is 01-01, set True, by default False
        """
        batchsize = kwargs.get("batchsize", 100)
        time_units = kwargs.get("time_units", self.time_unit) or [
            "1D"
        ]  # Default to ["1D"] if not specified or if time_units is None
        start0101_freq = kwargs.get("start0101_freq", False)

        variables = self.get_timeseries_cols()
        basins = self.camels_sites["basin_id"].values

        # Define the generator function for batching
        def data_generator(basins, batch_size):
            for i in range(0, len(basins), batch_size):
                yield basins[i : i + batch_size]

        for time_unit in time_units:
            if trange4cache is None:
                if time_unit != "3h":
                    trange4cache = ["1980-01-01", "2023-12-31"]
                else:
                    trange4cache = ["1980-01-01 01", "2023-12-31 22"]

            # Generate the time range specific to the time unit
            if start0101_freq:
                times = (
                    generate_start0101_time_range(
                        start_time=trange4cache[0],
                        end_time=trange4cache[-1],
                        freq=time_unit,
                    )
                    .strftime("%Y-%m-%d %H:%M:%S")
                    .tolist()
                )
            else:
                times = (
                    pd.date_range(
                        start=trange4cache[0], end=trange4cache[-1], freq=time_unit
                    )
                    .strftime("%Y-%m-%d %H:%M:%S")
                    .tolist()
                )
            # Retrieve the correct units information for this time unit
            unit_file = next(
                file
                for file in self.data_source_description["UNIT_FILES"]
                if time_unit in file
            )
            if "s3://" in unit_file:
                with conf.FS.open(unit_file, mode="rb") as fp:
                    units_info = json.load(fp)
            else:
                units_info = hydro_file.unserialize_json(unit_file)

            for basin_batch in data_generator(basins, batchsize):
                data = self.read_timeseries(
                    object_ids=basin_batch,
                    t_range_list=trange4cache,
                    relevant_cols=variables[
                        time_unit
                    ],  # Ensure we use the right columns for the time unit
                    time_units=[
                        time_unit
                    ],  # Pass the time unit to ensure correct data retrieval
                    start0101_freq=start0101_freq,
                )

                dataset = xr.Dataset(
                    data_vars={
                        variables[time_unit][i]: (
                            ["basin", "time"],
                            data[time_unit][:, :, i],
                            {"units": units_info[variables[time_unit][i]]},
                        )
                        for i in range(len(variables[time_unit]))
                    },
                    coords={
                        "basin": basin_batch,
                        "time": pd.to_datetime(times),
                    },
                )

                # Save the dataset to a NetCDF file for the current batch and time unit
                prefix_ = self._get_ts_file_prefix_(self.dataset_name, self.version)
                batch_file_path = os.path.join(
                    CACHE_DIR,
                    f"{prefix_}timeseries_{time_unit}_batch_{basin_batch[0]}_{basin_batch[-1]}.nc",
                )
                dataset.to_netcdf(batch_file_path)

                # Release memory by deleting the dataset
                del dataset
                del data

    def cache_xrdataset(self, t_range=None, time_units=None):
        """Save all data in a netcdf file in the cache directory"""
        self.cache_attributes_xrdataset()
        self.cache_timeseries_xrdataset(trange4cache=t_range, time_units=time_units)

    def read_ts_xrdataset(
        self,
        gage_id_lst: list = None,
        t_range: list = None,
        var_lst: list = None,
        **kwargs,
    ) -> dict:
        """
        Read time-series xarray dataset from multiple NetCDF files and organize them by time units.

        Parameters:
        ----------
        gage_id_lst: list
            List of gage IDs to select.
        t_range: list
            List of two elements [start_time, end_time] to select time range.
        var_lst: list
            List of variables to select.
        **kwargs
            Additional arguments.

        Returns:
        ----------
        dict: A dictionary where each key is a time unit and each value is an xarray.Dataset containing the selected gage IDs, time range, and variables.
        """
        dataset_name = self.dataset_name
        version = self.version
        time_units = kwargs.get("time_units", self.time_unit)
        if var_lst is None:
            return None

        # Initialize a dictionary to hold datasets for each time unit
        datasets_by_time_unit = {}

        prefix_ = self._get_ts_file_prefix_(dataset_name, version)

        for time_unit in time_units:
            # Collect batch files specific to the current time unit
            batch_files = self._get_batch_files(prefix_, time_unit)

            if not batch_files:
                # Cache the data if no batch files are found for the current time unit
                self.cache_timeseries_xrdataset(**kwargs)
                batch_files = self._get_batch_files(prefix_, time_unit)

            selected_datasets = []

            for batch_file in batch_files:
                ds = xr.open_dataset(batch_file)
                all_vars = ds.data_vars
                if any(var not in ds.variables for var in var_lst):
                    raise ValueError(f"var_lst must all be in {all_vars}")
                if valid_gage_ids := [
                    gid for gid in gage_id_lst if gid in ds["basin"].values
                ]:
                    ds_selected = ds[var_lst].sel(
                        basin=valid_gage_ids, time=slice(t_range[0], t_range[1])
                    )
                    selected_datasets.append(ds_selected)

                ds.close()  # Close the dataset to free memory

            # If any datasets were selected, concatenate them along the 'basin' dimension
            if selected_datasets:
                # NOTE: the chosen part must be sorted by basin, or there will be some negative sideeffect for continue usage of this repo
                datasets_by_time_unit[time_unit] = xr.concat(
                    selected_datasets, dim="basin"
                ).sortby("basin")
            else:
                datasets_by_time_unit[time_unit] = xr.Dataset()

        return datasets_by_time_unit

    def _get_ts_file_prefix_(self, dataset_name, version):
        prefix_ = "" if dataset_name is None else dataset_name + "_"
        # we add version for prefix_ as we will update the dataset iteratively
        prefix_ = prefix_ + f"{version}_" if version is not None else prefix_
        return prefix_

    def _get_batch_files(self, prefix_, time_unit):
        return [
            os.path.join(CACHE_DIR, f)
            for f in os.listdir(CACHE_DIR)
            if re.match(
                rf"^{prefix_}timeseries_{time_unit}_batch_[A-Za-z0-9_]+_[A-Za-z0-9_]+\.nc$",
                f,
            )
        ]

    def read_attr_xrdataset(self, gage_id_lst=None, var_lst=None, **kwargs):
        dataset_name = self.dataset_name

        prefix_ = "" if dataset_name is None else dataset_name + "_"
        if var_lst is None or len(var_lst) == 0:
            return None
        try:
            attr = xr.open_dataset(os.path.join(CACHE_DIR, f"{prefix_}attributes.nc"))
        except FileNotFoundError:
            self.cache_attributes_xrdataset()
            attr = xr.open_dataset(os.path.join(CACHE_DIR, f"{prefix_}attributes.nc"))
        return attr[var_lst].sel(basin=gage_id_lst)


def get_dor_values(gages: Gages, usgs_id) -> np.array:
    """
    get dor values from gages for the usgs_id-sites

    """

    assert all(x < y for x, y in zip(usgs_id, usgs_id[1:]))
    # mm/year 1-km grid,  megaliters total storage per sq km  (1 megaliters = 1,000,000 liters = 1,000 cubic meters)
    # attr_lst = ["RUNAVE7100", "STOR_NID_2009"]
    attr_lst = ["RUNAVE7100", "STOR_NOR_2009"]
    data_attr = gages.read_constant_cols(usgs_id, attr_lst)
    run_avg = data_attr[:, 0] * (10 ** (-3)) * (10**6)  # m^3 per year
    nor_storage = data_attr[:, 1] * 1000  # m^3
    return nor_storage / run_avg


def get_diversion(gages: Gages, usgs_id) -> np.array:
    diversion_strs = ["diversion", "divert"]
    assert all(x < y for x, y in zip(usgs_id, usgs_id[1:]))
    attr_lst = ["WR_REPORT_REMARKS", "SCREENING_COMMENTS"]
    data_attr = gages.read_attr_origin(usgs_id, attr_lst)
    diversion_strs_lower = [elem.lower() for elem in diversion_strs]
    data_attr0_lower = np.array(
        [elem.lower() if type(elem) == str else elem for elem in data_attr[0]]
    )
    data_attr1_lower = np.array(
        [elem.lower() if type(elem) == str else elem for elem in data_attr[1]]
    )
    data_attr_lower = np.vstack((data_attr0_lower, data_attr1_lower)).T
    diversions = [
        is_any_elem_in_a_lst(diversion_strs_lower, data_attr_lower[i], include=True)
        for i in range(len(usgs_id))
    ]
    return np.array(diversions)


def is_any_elem_in_a_lst(lst1, lst2, return_index=False, include=False):
    do_exist = False
    idx_lst = []
    for j in range(len(lst1)):
        if include:
            for lst2_elem in lst2:
                if lst1[j] in lst2_elem:
                    idx_lst.append(j)
                    do_exist = True
        elif lst1[j] in lst2:
            idx_lst.append(j)
            do_exist = True
    return (do_exist, idx_lst) if return_index else do_exist



"""
Author: Wenyu Ouyang
Date: 2023-11-03 09:16:41
LastEditTime: 2025-01-03 15:40:58
LastEditors: Wenyu Ouyang
Description: Reading GRDC data
FilePath: \hydrodatasource\hydrodatasource\reader\grdc.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import collections
import json
import os
import datetime
import logging
from pathlib import Path
import re
import numpy as np
import pandas as pd
import xarray as xr
import geopandas as gpd
from dateutil.parser import parse

from hydrodatasource.configs.config import CACHE_DIR, SETTING
from hydrodatasource.reader.data_source import HydroData


class Grdc(HydroData):
    """Reading GRDC streamflow data."""

    def __init__(self, data_path):
        self.data_source_dir = data_path
        self.data_source_description = self.set_data_source_describe()
        self.grdc_site_info = self.read_site_info()

    def get_name(self):
        return "GRDC"

    def set_data_source_describe(self):
        data_root_dir = self.data_source_dir
        continents = [
            "africa",
            "asia",
            "south_america",
            "north_america",
            "south_west_pacific",
            "europe",
        ]
        shape_dir = os.path.join(data_root_dir, "GRDC_Watersheds")
        shp_files = [f for f in os.listdir(shape_dir) if f.endswith(".shp")]

        if len(shp_files) != 1:
            raise ValueError(
                f"Expected one shapefile in {shape_dir}, found {len(shp_files)}"
            )

        shp_file_path = os.path.join(shape_dir, shp_files[0])
        data_description = {}
        for continent in continents:
            continent_dir = os.path.join(data_root_dir, continent)
            daily_dir = os.path.join(continent_dir, "daily")
            monthly_dir = os.path.join(continent_dir, "monthly")

            data_description[continent] = {
                "daily": daily_dir,
                "monthly": monthly_dir,
            }

        return collections.OrderedDict(
            DATA_DIR=data_root_dir,
            CONTINENT_DATA=data_description,
            # shapfile: https://mrb.grdc.bafg.de/
            BASINS_SHP_FILE=shp_file_path,
            # TODO: attribute: https://portal.grdc.bafg.de/applications/public.html?publicuser=PublicUser#dataDownload/StationCatalogue
        )

    def read_site_info(self):
        """Reads the shapefile and extracts the 'id' column as a list."""
        shp_file_path = self.data_source_description["BASINS_SHP_FILE"]
        gdf = gpd.read_file(shp_file_path)
        if "grdc_no" not in gdf.columns:
            raise ValueError(
                f"The shapefile {shp_file_path} does not contain an 'grdc_no' column."
            )
        gdf["grdc_no"] = gdf["grdc_no"].apply(
            lambda x: str(int(x)) if isinstance(x, float) else str(x)
        )
        assert all(x < y for x, y in zip(gdf["grdc_no"], gdf["grdc_no"][1:]))
        return gdf[["grdc_no", "area"]]

    def map_station_to_continent(self, station_id: str):
        """Maps a station ID to its corresponding continent based on rules."""
        if station_id.startswith("1"):
            return "africa"
        elif station_id.startswith("2"):
            return "asia"
        elif station_id.startswith("3"):
            return "south_america"
        elif station_id.startswith("4"):
            return "north_america"
        elif station_id.startswith("5"):
            return "south_west_pacific"
        elif station_id.startswith("6"):
            return "europe"
        else:
            raise ValueError(f"Invalid station ID: {station_id}")

    def read_grdc_daily_data(
        self,
        station_id: str,
        time_range: list,
        parameter: str = "Q",
        column: str = "streamflow",
    ):
        """read daily river discharge data from Global Runoff Data Centre (GRDC).

        Requires the GRDC daily data files in a local directory. The GRDC daily data
        files can be ordered at
        https://www.bafg.de/GRDC/EN/02_srvcs/21_tmsrs/riverdischarge_node.html

        Parameters
        ----------
            station_id: The station id to get. The station id can be found in the
                catalogues at
                https://www.bafg.de/GRDC/EN/02_srvcs/21_tmsrs/212_prjctlgs/project_catalogue_node.html
            time_range: A list of [start_time, end_time] in UTC and ISO format strings e.g.
                ['YYYY-MM-DDTHH:MM:SSZ', 'YYYY-MM-DDTHH:MM:SSZ'].
            parameter: optional. The parameter code to get, e.g. ('Q') discharge,
                cubic meters per second.
            column: optional. Name of column in dataframe. Default: "streamflow".

        Returns:
            grdc data in a dataframe and metadata.

        Examples:
            .. code-block:: python

                from ewatercycle.observation.grdc import get_grdc_data

                df, meta = get_grdc_data('6335020',
                                        ['2000-01-01T00:00Z', '2001-01-01T00:00Z'])
                df.describe()
                        streamflow
                count   4382.000000
                mean    2328.992469
                std     1190.181058
                min      881.000000
                25%     1550.000000
                50%     2000.000000
                75%     2730.000000
                max    11300.000000

                meta
                {'grdc_file_name': '/home/myusername/git/eWaterCycle/ewatercycle/6335020_Q_Day.Cmd.txt',
                'id_from_grdc': 6335020,
                'file_generation_date': '2019-03-27',
                'river_name': 'RHINE RIVER',
                'station_name': 'REES',
                'country_code': 'DE',
                'grdc_latitude_in_arc_degree': 51.756918,
                'grdc_longitude_in_arc_degree': 6.395395,
                'grdc_catchment_area_in_km2': 159300.0,
                'altitude_masl': 8.0,
                'dataSetContent': 'MEAN DAILY DISCHARGE (Q)',
                'units': 'm³/s',
                'time_series': '1814-11 - 2016-12',
                'no_of_years': 203,
                'last_update': '2018-05-24',
                'nrMeasurements': 'NA',
                'UserStartTime': '2000-01-01T00:00Z',
                'UserEndTime': '2001-01-01T00:00Z',
                'nrMissingData': 0}
        """  # noqa: E501
        continent = self.map_station_to_continent(station_id)
        data_path = self.data_source_description["CONTINENT_DATA"][continent]["daily"]

        if not os.path.exists(data_path):
            raise ValueError(f"The GRDC daily directory {data_path} does not exist!")

        # Read the raw data
        start_time, end_time = time_range
        raw_file = os.path.join(data_path, f"{station_id}_{parameter}_Day.Cmd.txt")
        if not os.path.exists(raw_file):
            raise ValueError(f"The GRDC file {raw_file} does not exist!")

        # Convert the raw data to an xarray
        metadata, df = _grdc_read(
            raw_file,
            start=_get_time(start_time).date(),
            end=_get_time(end_time).date(),
            column=column,
        )

        # Add start/end_time to metadata
        metadata["UserStartTime"] = start_time
        metadata["UserEndTime"] = end_time

        # Add number of missing data to metadata
        metadata["nrMissingData"] = _count_missing_data(df, column)

        # Show info about data
        _log_metadata(metadata)

        return df, metadata

    def cache_grdc_daily(self, station_ids=None, time_range=None, batch_size=1000):
        """
        Save GRDC daily data to a NetCDF file.

        Parameters
        ----------
        station_ids: list of str
            List of station IDs to read data for.
        time_range: list of str
            List of [start_time, end_time] in UTC and ISO format strings e.g.
            ['YYYY-MM-DDTHH:MM:SSZ', 'YYYY-MM-DDTHH:MM:SSZ'].
        batch_size: int
            Number of stations to process in each batch
        """
        if time_range is None:
            time_range = ["1980-01-01", "2023-12-31"]
        start_date, end_date = time_range
        catalogue = self.grdc_site_info
        if station_ids is None:
            station_ids = catalogue["grdc_no"].tolist()
        catalogue = catalogue[catalogue["grdc_no"].isin(station_ids)]

        def data_generator(station_ids, batch_size):
            for i in range(0, len(station_ids), batch_size):
                yield station_ids[i : i + batch_size]

        # Split station IDs into batches
        station_batches = list(data_generator(station_ids, batch_size))

        # Loop over each station in the catalogue
        for batch_index, batch in enumerate(station_batches):
            # Create empty lists to store data and metadata
            data_list = []
            meta_list = []
            for station_id in batch:
                try:
                    st = datetime.datetime.strptime(start_date, "%Y-%m-%d").strftime(
                        "%Y-%m-%dT%H:%M:%SZ"
                    )
                    et = datetime.datetime.strptime(end_date, "%Y-%m-%d").strftime(
                        "%Y-%m-%dT%H:%M:%SZ"
                    )
                    df, meta = self.read_grdc_daily_data(str(station_id), [st, et])
                except Exception as e:
                    print(f"Error reading data for station {station_id}: {e}")
                    # Create an empty DataFrame with the same structure
                    df = pd.DataFrame(
                        columns=["streamflow"],
                        index=pd.date_range(start=start_date, end=end_date),
                    )
                    df["streamflow"] = float("nan")
                    meta = {
                        "grdc_file_name": "",
                        "id_from_grdc": station_id,
                        "river_name": "",
                        "station_name": "",
                        "country_code": "",
                        "grdc_latitude_in_arc_degree": float("nan"),
                        "grdc_longitude_in_arc_degree": float("nan"),
                        "grdc_catchment_area_in_km2": float("nan"),
                        "altitude_masl": float("nan"),
                        "dataSetContent": "",
                        "units": "m³/s",
                        "time_series": "",
                        "no_of_years": 0,
                        "last_update": "",
                        "nrMeasurements": "NA",
                        "UserStartTime": start_date,
                        "UserEndTime": end_date,
                        "nrMissingData": _count_missing_data(df, "streamflow"),
                    }

                coords_dict = {"time": df.index, "station": [station_id]}
                da = xr.DataArray(
                    data=df["streamflow"].values.reshape(-1, 1),
                    coords=coords_dict,
                    dims=["time", "station"],
                    name="streamflow",
                    attrs={"units": meta.get("units", "unknown")},
                )
                data_list.append(da)

                # Append metadata
                meta_list.append(meta)

            # Concatenate all DataArrays along the 'station' dimension
            ds = xr.concat(data_list, dim="station").sortby("station")

            # Assign attributes
            ds.attrs["description"] = "Daily river discharge"
            ds.station.attrs["description"] = "GRDC station number"

            # Write the xarray Dataset to a NetCDF file
            batch_file_path = os.path.join(
                CACHE_DIR,
                f"grdc_daily_streamflow_data_batch_{batch[0]}_{batch[-1]}.nc",
            )
            ds.to_netcdf(batch_file_path)
            print(
                f"Batch {batch_index + 1}/{len(station_batches)} NetCDF file created successfully!"
            )
            # Convert meta_list to a serializable format
            serializable_meta_list = _convert_to_serializable(meta_list)
            # Write the metadata to a JSON file for the current batch
            meta_file_path = os.path.join(
                CACHE_DIR,
                f"grdc_daily_streamflow_metadata_batch_{batch[0]}_{batch[-1]}.json",
            )
            with open(meta_file_path, "w") as meta_file:
                json.dump(serializable_meta_list, meta_file)
            print(
                f"Batch {batch_index + 1}/{len(station_batches)} metadata file created successfully!"
            )

            # Release memory by deleting the dataset
            del ds
            del data_list
            del meta_list

    def read_streamflow_xrdataset(
        self,
        station_id_lst: list = None,
        time_range: list = None,
        **kwargs,
    ) -> dict:
        """
        Read GRDC daily data from multiple NetCDF files and organize them by station IDs.

        Parameters
        ----------
        station_id_lst : list
            List of station IDs to select.
        time_range : list
            List of two elements [start_time, end_time] to select time range.
        **kwargs: Additional arguments.

        Returns
        ----------
        dict: A dictionary where each key is a station ID and each value is an xarray.Dataset containing the selected station IDs, time range, and variable.
        """
        if station_id_lst is None or time_range is None:
            return None

        # Initialize a dictionary to hold datasets for each station ID
        datasets_by_station_id = {}

        # Collect batch files
        batch_files = [
            os.path.join(CACHE_DIR, f)
            for f in os.listdir(CACHE_DIR)
            if re.match(
                rf"^grdc_daily_streamflow_data_batch_[A-Za-z0-9_]+_[A-Za-z0-9_]+\.nc$",
                f,
            )
        ]

        if not batch_files:
            # Cache the data if no batch files are found
            self.cache_grdc_daily(
                station_ids=station_id_lst, time_range=time_range, **kwargs
            )
            batch_files = [
                os.path.join(CACHE_DIR, f)
                for f in os.listdir(CACHE_DIR)
                if re.match(
                    rf"^grdc_daily_streamflow_data_batch_[A-Za-z0-9_]+_[A-Za-z0-9_]+\.nc$",
                    f,
                )
            ]

        selected_datasets = []

        for batch_file in batch_files:
            ds = xr.open_dataset(batch_file)
            if valid_station_ids := [
                sid for sid in station_id_lst if sid in ds["station"].values
            ]:
                ds_selected = ds[["streamflow"]].sel(
                    station=valid_station_ids, time=slice(time_range[0], time_range[1])
                )
                selected_datasets.append(ds_selected)

            ds.close()  # Close the dataset to free memory

        # If any datasets were selected, concatenate them along the 'station' dimension
        if selected_datasets:
            datasets_by_station_id = xr.concat(selected_datasets, dim="station").sortby(
                "station"
            )
        else:
            datasets_by_station_id = xr.Dataset()

        return datasets_by_station_id


def _get_time(time_iso: str) -> datetime.datetime:
    """Return a datetime in UTC.
    Convert a date string in ISO format to a datetime
    and check if it is in UTC.
    """
    time = parse(time_iso)
    if time.tzname() != "UTC":
        raise ValueError(
            "The time is not in UTC. The ISO format for a UTC time "
            "is 'YYYY-MM-DDTHH:MM:SSZ'"
        )
    return time


def _grdc_metadata_reader(grdc_station_path, all_lines):
    """
    Initiating a dictionary that will contain all GRDC attributes.
    This function is based on earlier work by Rolf Hut.
    https://github.com/RolfHut/GRDC2NetCDF/blob/master/GRDC2NetCDF.py
    DOI: 10.5281/zenodo.19695
    that function was based on earlier work by Edwin Sutanudjaja from Utrecht University.
    https://github.com/edwinkost/discharge_analysis_IWMI
    Modified by Susan Branchett
    """

    # split the content of the file into several lines
    all_lines = all_lines.replace("\r", "")
    all_lines = all_lines.split("\n")

    # get grdc ids (from files) and check their consistency with their
    # file names
    id_from_file_name = int(
        os.path.basename(grdc_station_path).split(".")[0].split("_")[0]
    )
    id_from_grdc = None
    if id_from_file_name == int(all_lines[8].split(":")[1].strip()):
        id_from_grdc = int(all_lines[8].split(":")[1].strip())
    else:
        print(
            f"GRDC station {id_from_file_name} ({str(grdc_station_path)}) is NOT used."
        )

    attribute_grdc = {}
    if id_from_grdc is not None:
        attribute_grdc["grdc_file_name"] = str(grdc_station_path)
        attribute_grdc["id_from_grdc"] = id_from_grdc

        try:
            attribute_grdc["file_generation_date"] = str(
                all_lines[6].split(":")[1].strip()
            )
        except (IndexError, ValueError):
            attribute_grdc["file_generation_date"] = "NA"

        try:
            attribute_grdc["river_name"] = str(all_lines[9].split(":")[1].strip())
        except (IndexError, ValueError):
            attribute_grdc["river_name"] = "NA"

        try:
            attribute_grdc["station_name"] = str(all_lines[10].split(":")[1].strip())
        except (IndexError, ValueError):
            attribute_grdc["station_name"] = "NA"

        try:
            attribute_grdc["country_code"] = str(all_lines[11].split(":")[1].strip())
        except (IndexError, ValueError):
            attribute_grdc["country_code"] = "NA"

        try:
            attribute_grdc["grdc_latitude_in_arc_degree"] = float(
                all_lines[12].split(":")[1].strip()
            )
        except (IndexError, ValueError):
            attribute_grdc["grdc_latitude_in_arc_degree"] = "NA"

        try:
            attribute_grdc["grdc_longitude_in_arc_degree"] = float(
                all_lines[13].split(":")[1].strip()
            )
        except (IndexError, ValueError):
            attribute_grdc["grdc_longitude_in_arc_degree"] = "NA"

        try:
            attribute_grdc["grdc_catchment_area_in_km2"] = float(
                all_lines[14].split(":")[1].strip()
            )
            if attribute_grdc["grdc_catchment_area_in_km2"] <= 0.0:
                attribute_grdc["grdc_catchment_area_in_km2"] = "NA"
        except (IndexError, ValueError):
            attribute_grdc["grdc_catchment_area_in_km2"] = "NA"

        try:
            attribute_grdc["altitude_masl"] = float(all_lines[15].split(":")[1].strip())
        except (IndexError, ValueError):
            attribute_grdc["altitude_masl"] = "NA"

        try:
            attribute_grdc["dataSetContent"] = str(all_lines[21].split(":")[1].strip())
        except (IndexError, ValueError):
            attribute_grdc["dataSetContent"] = "NA"

        try:
            attribute_grdc["units"] = str(all_lines[23].split(":")[1].strip())
        except (IndexError, ValueError):
            attribute_grdc["units"] = "NA"

        try:
            attribute_grdc["time_series"] = str(all_lines[24].split(":")[1].strip())
        except (IndexError, ValueError):
            attribute_grdc["time_series"] = "NA"

        try:
            attribute_grdc["no_of_years"] = int(all_lines[25].split(":")[1].strip())
        except (IndexError, ValueError):
            attribute_grdc["no_of_years"] = "NA"

        try:
            attribute_grdc["last_update"] = str(all_lines[26].split(":")[1].strip())
        except (IndexError, ValueError):
            attribute_grdc["last_update"] = "NA"

        try:
            attribute_grdc["nrMeasurements"] = int(
                str(all_lines[34].split(":")[1].strip())
            )
        except (IndexError, ValueError):
            attribute_grdc["nrMeasurements"] = "NA"

    return attribute_grdc


def _grdc_read(grdc_station_file, start, end, column):
    grdc_station_path = Path(grdc_station_file)
    with grdc_station_path.open("r", encoding="cp1252", errors="ignore") as file:
        data = file.read()

    metadata = _grdc_metadata_reader(grdc_station_path, data)

    all_lines = data.split("\n")
    header = next(
        (i + 1 for i, line in enumerate(all_lines) if line.startswith("# DATA")),
        0,
    )
    # Import GRDC data into dataframe and modify dataframe format
    grdc_data = pd.read_csv(
        grdc_station_path,
        encoding="cp1252",
        skiprows=header,
        delimiter=";",
        parse_dates=["YYYY-MM-DD"],
        na_values="-999",
    )
    grdc_station_df = pd.DataFrame(
        {column: grdc_data[" Value"].array},
        index=grdc_data["YYYY-MM-DD"].array,
    )
    grdc_station_df.index.rename("time", inplace=True)

    # Create a continuous date range based on the given start and end dates
    full_date_range = pd.date_range(start=start, end=end)
    full_df = pd.DataFrame(index=full_date_range)
    full_df.index.rename("time", inplace=True)

    # Merge the two dataframes, so the dates without data will have NaN values
    merged_df = full_df.merge(
        grdc_station_df, left_index=True, right_index=True, how="left"
    )

    return metadata, merged_df


def _count_missing_data(df, column):
    """Return number of missing data."""
    return df[column].isna().sum()


def _log_metadata(metadata):
    """Print some information about data."""
    logger = logging.getLogger(__name__)
    coords = (
        metadata["grdc_latitude_in_arc_degree"],
        metadata["grdc_longitude_in_arc_degree"],
    )
    message = (
        f"GRDC station {metadata['id_from_grdc']} is selected. "
        f"The river name is: {metadata['river_name']}."
        f"The coordinates are: {coords}."
        f"The catchment area in km2 is: {metadata['grdc_catchment_area_in_km2']}. "
        f"There are {metadata['nrMissingData']} missing values during "
        f"{metadata['UserStartTime']}_{metadata['UserEndTime']} at this station. "
        f"See the metadata for more information."
    )
    logger.info("%s", message)


def _convert_to_serializable(obj):
    if isinstance(obj, np.int64):
        return int(obj)
    elif isinstance(obj, np.float64):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, dict):
        return {k: _convert_to_serializable(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_convert_to_serializable(i) for i in obj]
    return obj


if __name__ == "__main__":
    data_dir = os.path.join(SETTING["local_data_path"]["datasets-origin"], "GRDC")
    grdc = Grdc(data_dir)
    # grdc.cache_grdc_daily(["1107700", "4101200"], ["1990-10-01", "2000-10-01"])
    grdc.cache_grdc_daily()
    grdc.read_streamflow_xrdataset(["1107700", "4101200"], ["1990-10-01", "2000-10-01"])



"""
Author: Wenyu Ouyang
Date: 2024-07-06 19:20:59
LastEditTime: 2025-11-04 10:17:39
LastEditors: Wenyu Ouyang
Description: Test funcs for data source
FilePath: \hydrodatasource\tests\test_data_source.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pytest

pytestmark = pytest.mark.internal_data
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from hydrodatasource.configs.config import CACHE_DIR, SETTING
from hydrodatasource.reader.data_source import (
    SelfMadeForecastDataset,
    SelfMadeHydroDataset,
)


@pytest.fixture
def one_hour_dataset():
    # local
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]
    # minio
    # selfmadehydrodataset_path = "s3://basins-interim"
    return SelfMadeHydroDataset(
        data_path=selfmadehydrodataset_path, dataset_name="FDSources", time_unit=["1h"]
    )


@pytest.fixture
def three_hour_dataset():
    # local
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]
    # minio
    # selfmadehydrodataset_path = "s3://basins-interim"
    return SelfMadeHydroDataset(
        data_path=selfmadehydrodataset_path, dataset_name="FDSources", time_unit=["3h"]
    )


@pytest.fixture
def one_day_dataset():
    # local
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]
    # minio
    # selfmadehydrodataset_path = "s3://basins-interim"
    return SelfMadeHydroDataset(
        data_path=selfmadehydrodataset_path, dataset_name="FDSources"
    )


@pytest.fixture
def eight_day_dataset():
    # local
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]
    # minio
    # selfmadehydrodataset_path = "s3://basins-interim"
    return SelfMadeHydroDataset(
        data_path=selfmadehydrodataset_path,
        dataset_name="FDSources",
        time_unit=["8D"],
    )


def test_selfmadehydrodataset_get_name(one_day_dataset):
    assert one_day_dataset.get_name() == "SelfMadeHydroDataset"


def test_selfmadehydrodataset_streamflow_unit(one_day_dataset):
    assert one_day_dataset.streamflow_unit == {"1D": "mm/d"}


def test_selfmadehydrodataset_read_site_info(one_day_dataset):
    site_info = one_day_dataset.read_site_info()
    assert isinstance(site_info, pd.DataFrame)


def test_selfmadehydrodataset_read_object_ids(one_day_dataset):
    object_ids = one_day_dataset.read_object_ids()
    assert isinstance(object_ids, np.ndarray)


def test_selfmadehydrodataset_read_tsdata(one_day_dataset):
    object_ids = one_day_dataset.read_object_ids()
    target_cols = one_day_dataset.read_timeseries(
        object_ids=object_ids[:5],
        t_range_list=["2020-01-01", "2020-12-31"],
        relevant_cols=["streamflow"],
        time_unit=["1D"],
    )
    assert isinstance(target_cols, dict)


def test_selfmadehydrodataset_read_attrdata(one_day_dataset):
    object_ids = one_day_dataset.read_object_ids()
    constant_cols = one_day_dataset.read_attributes(
        object_ids=object_ids[:5], constant_cols=["area"]
    )
    assert isinstance(constant_cols, np.ndarray)


def test_selfmadehydrodataset_get_attributes_cols(one_day_dataset):
    constant_cols = one_day_dataset.get_attributes_cols()
    assert isinstance(constant_cols, np.ndarray)


def test_selfmadehydrodataset_get_timeseries_cols(one_day_dataset):
    relevant_cols = one_day_dataset.get_timeseries_cols()
    assert isinstance(relevant_cols, dict)


def test_selfmadehydrodataset_cache_attributes_xrdataset(one_day_dataset):
    one_day_dataset.cache_attributes_xrdataset()
    assert os.path.exists(
        os.path.join(CACHE_DIR, f"{one_day_dataset.dataset_name}_attributes.nc")
    )


@pytest.mark.slow
def test_selfmadehydrodataset_cache_timeseries_xrdataset(
    one_day_dataset, three_hour_dataset, one_hour_dataset, eight_day_dataset
):
    # 8D
    eight_day_dataset.cache_timeseries_xrdataset(
        time_units=["8D"],
        trange4cache=["1980-01-01", "2023-12-31"],
        start0101_freq=True,
        batchsize=200,
    )
    # 1h
    one_hour_dataset.cache_timeseries_xrdataset(
        time_units=["1h"],
        trange4cache=["1980-01-01", "2023-12-31"],
    )
    # 3h
    three_hour_dataset.cache_timeseries_xrdataset(
        time_units=["3h"],
        trange4cache=["1980-01-01 01", "2023-12-31 22"],
    )
    # 1D
    one_day_dataset.cache_timeseries_xrdataset()


@pytest.mark.slow
def test_selfmadehydrodataset_cache_xrdataset(one_day_dataset):
    one_day_dataset.cache_xrdataset()


def test_selfmadehydrodataset_read_ts_xrdataset(
    one_day_dataset, three_hour_dataset, one_hour_dataset, eight_day_dataset
):
    # 1h
    xrdataset_dict = one_hour_dataset.read_ts_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        t_range=["2020-01-01", "2020-12-31 23"],
        var_lst=["streamflow"],
        time_units=["1h"],
    )
    target_cols = one_hour_dataset.read_timeseries(
        object_ids=["camels_01013500", "camels_01022500"],
        t_range_list=["2020-01-01", "2020-12-31 23"],
        relevant_cols=["streamflow"],
        time_units=["1h"],
    )
    assert isinstance(xrdataset_dict, dict)
    np.testing.assert_array_equal(
        xrdataset_dict["1h"]["streamflow"].values, target_cols["1h"][:, :, 0]
    )

    # 3h
    xrdataset_dict = three_hour_dataset.read_ts_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        t_range=["2020-01-01 00", "2020-12-31 00"],
        var_lst=["streamflow"],
        time_units=["3h"],
        start_hour_in_a_day=0,
    )
    target_cols = three_hour_dataset.read_timeseries(
        object_ids=["camels_01013500", "camels_01022500"],
        t_range_list=["2020-01-01 00", "2020-12-31 00"],
        relevant_cols=["streamflow"],
        time_units=["3h"],
        start_hour_in_a_day=0,
    )
    assert isinstance(xrdataset_dict, dict)
    np.testing.assert_array_equal(
        xrdataset_dict["3h"]["streamflow"].values, target_cols["3h"][:, :, 0]
    )

    # 1D
    xrdataset_dict = one_day_dataset.read_ts_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        t_range=["2020-01-01", "2020-12-31"],
        var_lst=["streamflow"],
        time_units=["1D"],
    )
    target_cols = one_day_dataset.read_timeseries(
        object_ids=["camels_01013500", "camels_01022500"],
        t_range_list=["2020-01-01", "2020-12-31"],
        relevant_cols=["streamflow"],
        time_unit=["1D"],
    )
    assert isinstance(xrdataset_dict, dict)
    np.testing.assert_array_equal(
        xrdataset_dict["1D"]["streamflow"].values, target_cols["1D"][:, :, 0]
    )

    # 8D
    xrdataset_dict = eight_day_dataset.read_ts_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        t_range=["2020-01-01", "2020-12-31"],
        var_lst=["ET_modis16a2006", "ET_modis16a2gf061"],
        time_units=["8D"],
    )
    target_cols = eight_day_dataset.read_timeseries(
        object_ids=["camels_01013500", "camels_01022500"],
        t_range_list=["2020-01-01", "2020-12-31"],
        relevant_cols=["ET_modis16a2006", "ET_modis16a2gf061"],
        time_units=["8D"],
    )
    assert isinstance(xrdataset_dict, dict)
    np.testing.assert_array_equal(
        xrdataset_dict["8D"]["ET_modis16a2006"].values, target_cols["8D"][:, :, 0]
    )


def test_selfmadehydrodataset_read_attr_xrdataset(one_day_dataset):
    xrdataset = one_day_dataset.read_attr_xrdataset(
        gage_id_lst=["camels_01013500", "camels_01022500"],
        var_lst=["area"],
    )
    assert isinstance(xrdataset, xr.Dataset)


def test_selfmadehydrodataset_read_area(one_day_dataset):
    area = one_day_dataset.read_area(gage_id_lst=["camels_01013500", "camels_01022500"])
    assert isinstance(area, xr.Dataset)


def test_selfmadehydrodataset_read_mean_prcp(one_day_dataset):
    mean_prcp = one_day_dataset.read_mean_prcp(
        gage_id_lst=["camels_01013500", "camels_01022500"]
    )
    assert isinstance(mean_prcp, xr.Dataset)
    assert mean_prcp["pre_mm_syr"].attrs["units"] == "mm/d"


def test_read_mean_prcp_mm_per_hour(one_day_dataset):
    mean_prcp = one_day_dataset.read_mean_prcp(
        gage_id_lst=["camels_01013500", "camels_01022500"], unit="mm/h"
    )
    mean_prcp_ = one_day_dataset.read_mean_prcp(
        gage_id_lst=["camels_01013500", "camels_01022500"]
    )
    assert isinstance(mean_prcp, xr.Dataset)
    assert mean_prcp["pre_mm_syr"].attrs["units"] == "mm/h"
    np.testing.assert_allclose(
        mean_prcp["pre_mm_syr"].values, mean_prcp_["pre_mm_syr"].values / 24
    )


def test_read_mean_prcp_mm_per_3hour(one_day_dataset):
    mean_prcp = one_day_dataset.read_mean_prcp(
        gage_id_lst=["camels_01013500", "camels_01022500"], unit="mm/3h"
    )
    mean_prcp_ = one_day_dataset.read_mean_prcp(
        gage_id_lst=["camels_01013500", "camels_01022500"]
    )
    assert isinstance(mean_prcp, xr.Dataset)
    assert mean_prcp["pre_mm_syr"].attrs["units"] == "mm/3h"
    np.testing.assert_allclose(
        mean_prcp["pre_mm_syr"].values, mean_prcp_["pre_mm_syr"].values / (24 / 3)
    )


def test_read_mean_prcp_mm_per_8day(one_day_dataset):
    mean_prcp = one_day_dataset.read_mean_prcp(
        gage_id_lst=["camels_01013500", "camels_01022500"], unit="mm/8d"
    )
    mean_prcp_ = one_day_dataset.read_mean_prcp(
        gage_id_lst=["camels_01013500", "camels_01022500"]
    )
    assert isinstance(mean_prcp, xr.Dataset)
    assert mean_prcp["pre_mm_syr"].attrs["units"] == "mm/8d"
    np.testing.assert_allclose(
        mean_prcp["pre_mm_syr"].values, mean_prcp_["pre_mm_syr"].values * 8
    )


def test_read_mean_prcp_invalid_unit(one_day_dataset):
    with pytest.raises(ValueError, match="unit must be one of"):
        one_day_dataset.read_mean_prcp(
            gage_id_lst=["camels_01013500", "camels_01022500"], unit="invalid_unit"
        )


@pytest.fixture
def one_day_forecast_dataset(tmpdir, mocker):
    # Create temporary directory structure for testing
    fdsources_dir = tmpdir.mkdir("FDSources")

    # Create timeseries directory with a subdirectory (e.g., for time unit)
    ts_dir = fdsources_dir.mkdir("timeseries")
    ts_1d_dir = ts_dir.mkdir("1D")  # Create a time unit subdirectory

    # Create forecast directory with subdirectories
    forecast_dir = fdsources_dir.mkdir("forecasts")
    basin_dir = forecast_dir.mkdir("basin_1")

    # Create other required directories
    attr_dir = fdsources_dir.mkdir("attributes")
    shape_dir = fdsources_dir.mkdir("shapes")

    # Create a dummy attributes.csv file
    attr_file = attr_dir.join("attributes.csv")
    attr_file.write("basin_id,area\nbasin_1,100\nbasin_2,200\n")

    # local
    selfmadehydrodataset_path = str(tmpdir)
    # minio
    # selfmadehydrodataset_path = "s3://basins-interim"
    return SelfMadeForecastDataset(
        data_path=selfmadehydrodataset_path, dataset_name="FDSources"
    )


def test_read_forecast_multiple_basins_all_exist(mocker, one_day_forecast_dataset):
    mocker.patch.object(
        one_day_forecast_dataset,
        "data_source_description",
        {"FORECAST_DIR": "/mock/forecast_dir"},
    )
    mocker.patch("os.path.exists", return_value=True)
    mock_data = pd.DataFrame(
        {
            "date": pd.date_range("2020-01-01", periods=3),
            "forecast_date": pd.date_range("2020-01-01", periods=3),
            "streamflow": [1.0, 2.0, 3.0],
        }
    )
    mocker.patch("pandas.read_csv", return_value=mock_data)
    result = one_day_forecast_dataset.read_forecast(
        object_ids=["basin_1", "basin_2"],
        t_range_list=["2020-01-01", "2020-01-03"],
        relevant_cols=["streamflow"],
    )
    assert isinstance(result, dict)
    assert "basin_1" in result and "basin_2" in result
    assert all(isinstance(df, pd.DataFrame) for df in result.values())


def test_start_hour_in_a_day_validation():
    """Test that start_hour_in_a_day validation works correctly for time interval format."""
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]

    # Test: Non-standard time_unit like 6h should be rejected at init level
    # (SelfMadeHydroDataset only supports '1h', '3h', '1D', '8D')
    with pytest.raises(ValueError) as excinfo:
        dataset_6h = SelfMadeHydroDataset(
            data_path=selfmadehydrodataset_path,
            dataset_name="FDSources",
            time_unit=["6h"],
        )
    assert "time_unit must be one of" in str(excinfo.value)

    # Test that the cache_timeseries_xrdataset validation catches unsupported intervals
    # when they are passed directly to the function
    dataset = SelfMadeHydroDataset(
        data_path=selfmadehydrodataset_path, dataset_name="FDSources", time_unit=["3h"]
    )
    dataset.trange4cache = None
    dataset.offset_to_utc = False

    # Try to use an unsupported interval through kwargs
    # This should be caught by our new validation logic
    with pytest.raises(ValueError) as excinfo:
        dataset.cache_timeseries_xrdataset(
            time_units=["6h"], start_hour_in_a_day=2, batchsize=10
        )
    assert "only '3h' sub-daily interval is supported" in str(excinfo.value)

    # Test invalid start_hour_in_a_day value (should be 0-23)
    with pytest.raises(ValueError) as excinfo:
        dataset.cache_timeseries_xrdataset(
            time_units=["3h"], start_hour_in_a_day=25, batchsize=10
        )
    assert "must be an integer between 0 and 23" in str(excinfo.value)

    # Test invalid start_hour_in_a_day type
    with pytest.raises(ValueError) as excinfo:
        dataset.cache_timeseries_xrdataset(
            time_units=["3h"], start_hour_in_a_day="02:00:00", batchsize=10
        )
    assert "must be an integer between 0 and 23" in str(excinfo.value)


def test_start_hour_in_a_day_time_range():
    """Test that start_hour_in_a_day correctly sets the time range."""
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]

    dataset = SelfMadeHydroDataset(
        data_path=selfmadehydrodataset_path, dataset_name="FDSources", time_unit=["3h"]
    )
    dataset.trange4cache = None
    dataset.offset_to_utc = False

    # Test with start_hour_in_a_day = 5
    # Expected: start with "05", end with "23" (05, 08, 11, 14, 17, 20, 23)
    try:
        dataset.cache_timeseries_xrdataset(
            time_units=["3h"], start_hour_in_a_day=5, batchsize=10
        )
    except Exception:
        # Ignore execution errors, we just want to check trange4cache
        pass

    # Check that trange4cache was set correctly
    assert dataset.trange4cache is not None
    assert "05" in dataset.trange4cache[0]
    assert "23" in dataset.trange4cache[1]


def test_start_hour_in_a_day_data_alignment(mocker):
    """Test that data alignment validation works correctly."""
    selfmadehydrodataset_path = SETTING["local_data_path"]["datasets-interim"]

    dataset = SelfMadeHydroDataset(
        data_path=selfmadehydrodataset_path, dataset_name="FDSources", time_unit=["3h"]
    )
    dataset.offset_to_utc = False

    # Mock data with hours starting at 00:00 (0, 3, 6, 9, 12, 15, 18, 21)
    mock_data = pd.DataFrame(
        {
            "time": pd.date_range("2020-01-01 00:00", periods=8, freq="3h"),
            "streamflow": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0],
        }
    )

    mocker.patch("pandas.read_csv", return_value=mock_data)

    # Test 1: Correct alignment - should not raise error
    # Data starts at 00:00, so start_hour_in_a_day should be 0
    try:
        result = dataset.read_timeseries(
            object_ids=["test_basin"],
            t_range_list=["2020-01-01 00", "2020-01-02 00"],
            relevant_cols=["streamflow"],
            time_units=["3h"],
            start_hour_in_a_day=0,
        )
        # If we get here, the validation passed
        assert "3h" in result
    except ValueError as e:
        pytest.fail(f"Should not raise error with correct alignment: {e}")

    # Test 2: Incorrect alignment - should raise error
    # Data starts at 00:00, but start_hour_in_a_day is set to 2
    with pytest.raises(ValueError) as excinfo:
        dataset.read_timeseries(
            object_ids=["test_basin"],
            t_range_list=["2020-01-01 00", "2020-01-02 00"],
            relevant_cols=["streamflow"],
            time_units=["3h"],
            start_hour_in_a_day=2,
        )

    # Check error message content
    assert "Data time alignment error" in str(excinfo.value)
    assert "Please set start_hour_in_a_day to 0" in str(excinfo.value)



"""
Author: Wenyu Ouyang
Date: 2023-11-01 08:58:50
LastEditTime: 2025-01-09 18:03:34
LastEditors: Wenyu Ouyang
Description: Test funcs for reader.py
FilePath: \hydrodatasource\tests\test_rsvr_inflow_reader.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import pytest
import pandas as pd
from hydrodatasource.reader.rsvr_inflow_reader import RsvrInflowReader


@pytest.fixture
def rsvr_inflow_reader(tmpdir):
    # Create a temporary directory for the test data
    data_folder = tmpdir.mkdir("data")
    # Create a sample reservoir info file
    rsvr_info_file = data_folder.join("rsvr_info.csv")
    rsvr_info_file.write("STCD\n0001\n0002\n")
    # Create sample inflow data files for reservoirs
    for rsvr_id in ["0001", "0002"]:
        rsvr_dir = data_folder.mkdir(rsvr_id)
        rsvr_data_file = rsvr_dir.join(f"{rsvr_id}_rsvr_data.csv")
        rsvr_data_file.write(
            "TM,STCD,INQ\n2023-01-01,0001,100.0\n2023-01-02,0001,110.0\n"
        )
    return RsvrInflowReader(data_folder)


def test_read_1rsvr_inflow(rsvr_inflow_reader):
    # Test reading inflow data for reservoir 0001
    rsvr_id = "0001"
    df = rsvr_inflow_reader.read_1rsvr_inflow(rsvr_id)
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert list(df.columns) == ["STCD", "INQ"]
    assert df.index.name == "TM"
    assert df.loc["2023-01-01", "INQ"] == 100.0
    assert df.loc["2023-01-02", "INQ"] == 110.0

    # Test reading inflow data for reservoir 0002
    rsvr_id = "0002"
    df = rsvr_inflow_reader.read_1rsvr_inflow(rsvr_id)
    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert list(df.columns) == ["STCD", "INQ"]
    assert df.index.name == "TM"
    assert df.loc["2023-01-01", "INQ"] == 100.0
    assert df.loc["2023-01-02", "INQ"] == 110.0



"""
Author: Wenyu Ouyang
Date: 2024-11-04 19:50:06
LastEditTime: 2025-11-04 10:16:31
LastEditors: Wenyu Ouyang
Description: test for gages
FilePath: \hydrodatasource\tests\test_gages.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pytest
import pandas as pd
from hydrodatasource.reader.gages import Gages
from hydrodatasource.configs.config import CACHE_DIR, SETTING

pytestmark = pytest.mark.internal_data


@pytest.fixture
def gages_dataset():
    # local
    gages_path = os.path.join(SETTING["local_data_path"]["datasets-interim"])
    # minio
    # gages_path = "s3://basins-interim"
    return Gages(data_path=gages_path, dataset_name="hydrodl-reservoir-jh-paper")


def test_gages_read_site_info(gages_dataset):
    site_info = gages_dataset.read_site_info()
    assert isinstance(site_info, pd.DataFrame)


def test_gages_cache_attributes_xrdataset(gages_dataset):
    gages_dataset.cache_attributes_xrdataset()
    assert os.path.exists(
        os.path.join(CACHE_DIR, "hydrodl-reservoir-jh-paper_attributes.nc")
    )
